"""Module for Serialization and Deserialization of a KNX Device Configuration ACK."""

from __future__ import annotations

from xknx.exceptions import CouldNotParseKNXIP

from .body import KNXIPBodyResponse
from .error_code import ErrorCode
from .knxip_enum import KNXIPServiceType


class DeviceConfigurationAck(KNXIPBodyResponse):
    """Representation of a KNX Device Configuration Ack."""

    SERVICE_TYPE = KNXIPServiceType.DEVICE_CONFIGURATION_ACK
    BODY_LENGTH = 4

    def __init__(
        self,
        communication_channel_id: int = 1,
        sequence_counter: int = 0,
        status_code: ErrorCode = ErrorCode.E_NO_ERROR,
    ) -> None:
        """Initialize DeviceConfigurationAck object."""
        self.communication_channel_id = communication_channel_id
        self.sequence_counter = sequence_counter
        self.status_code = status_code

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return DeviceConfigurationAck.BODY_LENGTH

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if raw[0] != DeviceConfigurationAck.BODY_LENGTH:  # structure_length field
            raise CouldNotParseKNXIP("DeviceConfigurationAck body has invalid length")
        if len(raw) != DeviceConfigurationAck.BODY_LENGTH:
            raise CouldNotParseKNXIP("DeviceConfigurationAck body has wrong length")
        self.communication_channel_id = raw[1]
        self.sequence_counter = raw[2]
        self.status_code = ErrorCode(raw[3])
        return DeviceConfigurationAck.BODY_LENGTH

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return bytes(
            (
                DeviceConfigurationAck.BODY_LENGTH,
                self.communication_channel_id,
                self.sequence_counter,
                self.status_code.value,
            )
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            "<DeviceConfigurationAck "
            f'communication_channel_id="{self.communication_channel_id}" '
            f'sequence_counter="{self.sequence_counter}" '
            f'status_code="{self.status_code}" />'
        )



"""Module for serialization and deserialization of KNX/IP Header."""

from __future__ import annotations

from typing import Final

from xknx.exceptions import CouldNotParseKNXIP, IncompleteKNXIPFrame

from .body import KNXIPBody
from .knxip_enum import KNXIPServiceType


class KNXIPHeader:
    """Class for serialization and deserialization of KNX/IP Header."""

    HEADERLENGTH: Final = 0x06
    PROTOCOLVERSION: Final = 0x10  # The only valid protocol version at this time is 1.0

    def __init__(self) -> None:
        """Initialize KNXIPHeader class."""
        self.service_type_ident = KNXIPServiceType.ROUTING_INDICATION
        self.total_length = 0  # to be set later

    def from_knx(self, data: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if len(data) < KNXIPHeader.HEADERLENGTH:
            raise IncompleteKNXIPFrame("wrong connection header length")
        if data[0] != KNXIPHeader.HEADERLENGTH:
            raise CouldNotParseKNXIP("wrong connection header length")
        # set immediately, as we need it for tcp stream parsing before raising exception
        self.total_length = data[4] * 256 + data[5]
        if data[1] != KNXIPHeader.PROTOCOLVERSION:
            raise CouldNotParseKNXIP("wrong protocol version")

        try:
            self.service_type_ident = KNXIPServiceType(data[2] * 256 + data[3])
        except ValueError:
            raise CouldNotParseKNXIP(
                f"KNXIPServiceType unknown: 0x{data[2:4].hex()}"
            ) from None
        return KNXIPHeader.HEADERLENGTH

    def set_length(self, body: KNXIPBody) -> None:
        """Set length of full KNX/IP packet from body + fixed header length."""
        if not isinstance(body, KNXIPBody):
            raise TypeError()
        self.total_length = KNXIPHeader.HEADERLENGTH + body.calculated_length()

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return (
            bytes((KNXIPHeader.HEADERLENGTH, KNXIPHeader.PROTOCOLVERSION))
            + self.service_type_ident.value.to_bytes(2, "big")
            + self.total_length.to_bytes(2, "big")
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            "<KNXIPHeader "
            f'HeaderLength="{KNXIPHeader.HEADERLENGTH}" '
            f'ProtocolVersion="{KNXIPHeader.PROTOCOLVERSION}" '
            f'KNXIPServiceType="{self.service_type_ident.name}" '
            f'TotalLength="{self.total_length}" />'
        )

    def __eq__(self, other: object) -> bool:
        """Equal operator."""
        return self.__dict__ == other.__dict__



"""
Module for Serialization and Deserialization of KNX Search Response Extended.

Search Requests Extended are used to search for KNX/IP devices within the network.
A search response extended contains all information of a found device (Name, serial number, supported features.).
It supports an array-style access to the DIBs (use classname as index). Every KNXnet/ip server shall send
a search response extended and one device supporting multiple KNX connections may send multiple search responses.

If there were any SRPs (Search Request Parameters) in the Search Request Extended and those were marked as mandatory
the server shall only reply to those requests if it can fulfill the requirements in the SRP.
"""

from __future__ import annotations

from .knxip_enum import KNXIPServiceType
from .search_response import SearchResponse


class SearchResponseExtended(SearchResponse):
    """Representation of a KNX Search Extended."""

    SERVICE_TYPE = KNXIPServiceType.SEARCH_RESPONSE_EXTENDED

    def __repr__(self) -> str:
        """Return object as readable string."""
        _dibs_str = ",\n".join(dib.__str__() for dib in self.dibs)
        return (
            "<SearchResponseExtended "
            f'control_endpoint="{self.control_endpoint}" '
            f'dibs="[\n{_dibs_str}\n]" />'
        )



"""Unit test for KNX/IP DisconnectRequest objects."""

import pytest

from xknx.exceptions import CouldNotParseKNXIP
from xknx.knxip import HPAI, DisconnectRequest, KNXIPFrame


class TestKNXIPDisconnectRequest:
    """Test class for KNX/IP DisconnectRequest objects."""

    def test_disconnect_request(self) -> None:
        """Test parsing and streaming DisconnectRequest KNX/IP packet."""
        raw = bytes.fromhex("06 10 02 09 00 10 15 00 08 01 C0 A8 C8 0C C3 B4")
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, DisconnectRequest)

        assert knxipframe.body.communication_channel_id == 21
        assert knxipframe.body.control_endpoint == HPAI(
            ip_addr="192.168.200.12", port=50100
        )

        disconnect_request = DisconnectRequest(
            communication_channel_id=21,
            control_endpoint=HPAI(ip_addr="192.168.200.12", port=50100),
        )
        knxipframe2 = KNXIPFrame.init_from_body(disconnect_request)

        assert knxipframe2.to_knx() == raw

    def test_from_knx_wrong_length(self) -> None:
        """Test parsing and streaming wrong DisconnectRequest."""
        raw = bytes((0x06, 0x10, 0x02, 0x09, 0x00, 0x10))
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)



"""
Module for managing a remote value on the KNX bus.

Remote value can be :
- a group address for writing a KNX value,
- a group address for reading a KNX value,
- or a group of both representing the same value.
"""

from __future__ import annotations

from abc import ABC
from collections.abc import Callable, Iterator
import logging
from typing import TYPE_CHECKING, Generic, TypeVar, Union

from xknx.dpt import DPTArray, DPTBase, DPTBinary
from xknx.exceptions import ConversionError, CouldNotParseTelegram
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.address import (
    DeviceGroupAddress,
    InternalGroupAddress,
    parse_device_group_address,
)
from xknx.telegram.apci import GroupValueResponse, GroupValueWrite

if TYPE_CHECKING:
    from xknx.telegram.address import DeviceAddressableType
    from xknx.xknx import XKNX

logger = logging.getLogger("xknx.log")

GroupAddressesType = Union[
    "DeviceAddressableType", list[Union["DeviceAddressableType", None]], None
]
ValueT = TypeVar("ValueT")

RVCallbackType = Callable[[ValueT], None]


class RemoteValue(ABC, Generic[ValueT]):
    """Class for managing remote knx value."""

    dpt_class: type[DPTBase] | None = None

    def __init__(
        self,
        xknx: XKNX,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        sync_state: None | bool | int | float | str = None,
        device_name: str | None = None,
        feature_name: str | None = None,
        after_update_cb: RVCallbackType[ValueT] | None = None,
    ) -> None:
        """Initialize RemoteValue class."""
        self.xknx: XKNX = xknx
        self.passive_group_addresses: list[DeviceGroupAddress] = []

        def unpack_group_addresses(
            addresses: GroupAddressesType,
        ) -> DeviceGroupAddress | None:
            """Parse group addresses and assign passive addresses when given."""
            if addresses is None:
                return None
            if not isinstance(addresses, list):
                return parse_device_group_address(addresses)
            if not addresses:  # empty list
                return None
            active = addresses[0]
            passive = [
                parse_device_group_address(addr)
                for addr in addresses[1:]
                if addr is not None
            ]
            self.passive_group_addresses.extend(passive)
            return parse_device_group_address(active) if active is not None else None

        self.group_address = unpack_group_addresses(group_address)
        self.group_address_state = unpack_group_addresses(group_address_state)

        self.device_name: str = "Unknown" if device_name is None else device_name
        self.feature_name: str = "Unknown" if feature_name is None else feature_name
        self._value: ValueT | None = None
        self._payload: DPTArray | DPTBinary | None = None
        self.telegram: Telegram | None = None
        self.after_update_cb: RVCallbackType[ValueT] | None = after_update_cb
        self._sync_state = sync_state

    def group_addresses(self) -> Iterator[DeviceGroupAddress]:
        """Return all configured group addresses of this RemoteValue."""
        if self.group_address is not None:
            yield self.group_address
        if self.group_address_state is not None:
            yield self.group_address_state
        yield from self.passive_group_addresses

    def register_state_updater(self) -> None:
        """Register RemoteValue for state updates."""
        sync_state = (
            self._sync_state
            if self._sync_state is not None
            else self.xknx.state_updater.default_use_updater
        )
        if sync_state and self.group_address_state:
            self.xknx.state_updater.register_remote_value(
                self, tracker_options=sync_state
            )

    def unregister_state_updater(self) -> None:
        """Unregister RemoteValue from state updates."""
        try:
            self.xknx.state_updater.unregister_remote_value(self)
        except KeyError:
            # KeyError if it was never added to StateUpdater
            pass

    @property
    def value(self) -> ValueT | None:
        """Get current value."""
        return self._value

    @value.setter
    def value(self, value: ValueT | None) -> None:
        """Set new value without creating a Telegram or calling after_update_cb. Raises ConversionError on invalid value."""
        if value is not None:
            # raises ConversionError on invalid value
            self._payload = self.to_knx(value)
        else:
            self._payload = None
        self._value = value

    @property
    def last_payload(self) -> DPTArray | DPTBinary | None:
        """Get last payload."""
        return self._payload

    def update_value(self, value: ValueT) -> None:
        """Set new value without creating a Telegram. Calls after_update_cb. Raises ConversionError on invalid value."""
        self.value = value
        if self.after_update_cb is not None:
            self.after_update_cb(value)

    @property
    def initialized(self) -> bool:
        """Evaluate if remote value is initialized with group address."""
        return bool(
            self.group_address_state
            or self.group_address
            or self.passive_group_addresses
        )

    @property
    def readable(self) -> bool:
        """Evaluate if remote value should be read from bus."""
        return bool(self.group_address_state)

    @property
    def writable(self) -> bool:
        """Evaluate if remote value has a group_address set."""
        return bool(self.group_address)

    def from_knx(self, payload: DPTArray | DPTBinary) -> ValueT:
        """Convert current payload to value - to be implemented in derived class when `dpt_class` can't be used."""
        if self.dpt_class is None:
            raise NotImplementedError(
                "Either `dpt_class` must be set or `from_knx` must be implemented"
            )
        return self.dpt_class.from_knx(payload)  # type: ignore[no-any-return]

    def to_knx(self, value: ValueT) -> DPTArray | DPTBinary:
        """Convert value to payload - to be implemented in derived class when `dpt_class` can't be used."""
        if self.dpt_class is None:
            raise NotImplementedError(
                "Either `dpt_class` must be set or `to_knx` must be implemented"
            )
        return self.dpt_class.to_knx(value)

    def process(self, telegram: Telegram, always_callback: bool = False) -> bool:
        """Process incoming or outgoing telegram."""
        if (
            not isinstance(
                telegram.destination_address, GroupAddress | InternalGroupAddress
            )
            or telegram.destination_address not in self.group_addresses()
        ):
            return False
        if not isinstance(telegram.payload, GroupValueWrite | GroupValueResponse):
            raise CouldNotParseTelegram(
                "payload not a GroupValueWrite or GroupValueResponse",
                payload=str(telegram.payload),
                destination_address=str(telegram.destination_address),
                source_address=str(telegram.source_address),
                device_name=self.device_name,
                feature_name=self.feature_name,
            )

        try:
            decoded_payload: ValueT
            if (
                telegram.decoded_data is not None
                and telegram.decoded_data.transcoder is self.dpt_class
            ):
                decoded_payload = telegram.decoded_data.value  # type: ignore[assignment]
            else:
                decoded_payload = self.from_knx(telegram.payload.value)
        except (ConversionError, CouldNotParseTelegram) as err:
            logger.warning(
                "Can not process %s for %s - %s: %s",
                telegram,
                self.device_name,
                self.feature_name,
                err,
            )
            return False
        self._payload = telegram.payload.value
        self.xknx.state_updater.update_received(self)
        if self._value is None or always_callback or self._value != decoded_payload:
            self._value = decoded_payload
            self.telegram = telegram
            if self.after_update_cb is not None:
                self.after_update_cb(decoded_payload)
        return True

    def send_raw(self, payload: DPTArray | DPTBinary, response: bool = False) -> None:
        """Send payload as telegram to KNX bus."""
        if self.group_address is None:
            logger.warning(
                "Attempted to set payload for non-writable device: %s - %s (payload: %s)",
                self.device_name,
                self.feature_name,
                payload,
            )
            return
        telegram = Telegram(
            destination_address=self.group_address,
            payload=(
                GroupValueResponse(payload) if response else GroupValueWrite(payload)
            ),
            source_address=self.xknx.current_address,
        )
        self.xknx.telegrams.put_nowait(telegram)

    def set(self, value: ValueT, response: bool = False) -> None:
        """Set new value."""
        if not self.writable:
            logger.warning(
                "Attempted to set value for non-writable device: %s - %s (value: %s)",
                self.device_name,
                self.feature_name,
                value,
            )
            return
        payload = self.to_knx(value)
        # self._value is set and after_update_cb() called when the outgoing telegram is processed.
        self.send_raw(payload, response)

    def respond(self) -> None:
        """Send current payload as GroupValueResponse telegram to KNX bus."""
        if self._value is None:
            return
        payload = self.to_knx(self._value)
        self.send_raw(payload, response=True)

    async def read_state(self, wait_for_result: bool = False) -> None:
        """Send GroupValueRead telegram for state address to KNX bus."""
        if self.group_address_state is not None:
            # pylint: disable=import-outside-toplevel
            # TODO: send a ReadRequest and start a timeout from here instead of ValueReader
            #       cancel timeout form process(); delete ValueReader
            from xknx.core import ValueReader

            value_reader = ValueReader(self.xknx, self.group_address_state)
            if wait_for_result:
                if await value_reader.read() is None:
                    logger.warning(
                        "Could not sync group address '%s' (%s - %s)",
                        self.group_address_state,
                        self.device_name,
                        self.feature_name,
                    )
            else:
                value_reader.send_group_read()

    @property
    def unit_of_measurement(self) -> str | None:
        """Return the unit of measurement."""
        return None

    def group_addr_str(self) -> str:
        """Return object as readable string."""
        return (
            f"<{self.group_address}, "
            f"{self.group_address_state}, "
            f"{list(map(str, self.passive_group_addresses))}, "
            f"{self.value!r} />"
        )

    def __str__(self) -> str:
        """Return object as string representation."""
        return (
            f"<{self.__class__.__name__} "
            f'device_name="{self.device_name}" '
            f'feature_name="{self.feature_name}" '
            f"{self.group_addr_str()} />"
        )

    def __eq__(self, other: object) -> bool:
        """Equal operator."""
        for key, value in self.__dict__.items():
            if key == "after_update_cb":
                continue
            if key not in other.__dict__:
                return False
            if other.__dict__[key] != value:
                return False
        for key in other.__dict__:
            if key == "after_update_cb":
                continue
            if key not in self.__dict__:
                return False
        return True



"""Unit test for KNX/IP Authenticate Request/Response."""

from unittest.mock import Mock, patch

from xknx.io.request_response import Authenticate
from xknx.knxip import (
    HPAI,
    ConnectionStateRequest,
    KNXIPFrame,
    SessionAuthenticate,
    SessionStatus,
)
from xknx.knxip.knxip_enum import SecureSessionStatusCode


class TestAuthenticate:
    """Test class for xknx/io/Authenticate objects."""

    async def test_authenticate(self) -> None:
        """Test authenticating to secure KNX device."""
        transport_mock = Mock()
        user_id = 123
        mac = bytes(16)
        authenticate = Authenticate(
            transport_mock,
            user_id=user_id,
            message_authentication_code=mac,
        )
        authenticate.timeout_in_seconds = 0

        assert authenticate.awaited_response_class == SessionStatus

        # Expected KNX/IP-Frame:
        exp_knxipframe = KNXIPFrame.init_from_body(
            SessionAuthenticate(
                user_id=user_id,
                message_authentication_code=mac,
            )
        )

        await authenticate.start()
        transport_mock.send.assert_called_with(exp_knxipframe)

        # Response KNX/IP-Frame with wrong type
        wrong_knxipframe = KNXIPFrame.init_from_body(ConnectionStateRequest())
        with patch("logging.Logger.warning") as mock_warning:
            authenticate.response_rec_callback(wrong_knxipframe, HPAI(), None)
            mock_warning.assert_called_with("Could not understand knxipframe")
            assert authenticate.success is False

        # Correct Response KNX/IP-Frame:
        res_knxipframe = KNXIPFrame.init_from_body(SessionStatus())
        res_knxipframe.body.status = (
            SecureSessionStatusCode.STATUS_AUTHENTICATION_SUCCESS
        )
        authenticate.response_rec_callback(res_knxipframe, HPAI(), None)
        assert authenticate.success is True
        assert (
            authenticate.response.status
            == SecureSessionStatusCode.STATUS_AUTHENTICATION_SUCCESS
        )



"""Unit tests for exceptions."""

import pytest

from xknx.exceptions import (
    ConversionError,
    CouldNotParseAddress,
    CouldNotParseKNXIP,
    CouldNotParseTelegram,
    DeviceIllegalValue,
    XKNXException,
)


@pytest.mark.parametrize(
    "base,equal,diff",
    [
        (
            ConversionError("desc1"),
            ConversionError("desc1"),
            ConversionError("desc2"),
        ),
        (
            CouldNotParseAddress(123),
            CouldNotParseAddress(123),
            CouldNotParseAddress(321),
        ),
        (
            CouldNotParseKNXIP("desc1"),
            CouldNotParseKNXIP("desc1"),
            CouldNotParseKNXIP("desc2"),
        ),
        (
            CouldNotParseTelegram("desc", arg1=1, arg2=2),
            CouldNotParseTelegram("desc", arg1=1, arg2=2),
            CouldNotParseTelegram("desc", arg1=2, arg2=1),
        ),
        (
            DeviceIllegalValue("value1", "desc"),
            DeviceIllegalValue("value1", "desc"),
            DeviceIllegalValue("value1", "desc2"),
        ),
        (
            XKNXException("desc1"),
            XKNXException("desc1"),
            XKNXException("desc2"),
        ),
    ],
)
def test_exceptions(
    base: XKNXException,
    equal: XKNXException,
    diff: XKNXException,
) -> None:
    """Test hashability and repr of exceptions."""
    assert hash(base) == hash(equal)
    assert hash(base) != hash(diff)
    assert base == equal
    assert base != diff
    assert repr(base) == repr(equal)
    assert repr(base) != repr(diff)



"""Implementation of the KNX 235 data point type."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass

from .dpt import DPTComplex, DPTComplexData
from .dpt_5 import DPTTariff
from .dpt_13 import DPTActiveEnergy
from .payload import DPTArray, DPTBinary


@dataclass(slots=True)
class TariffActiveEnergy(DPTComplexData):
    """
    Representation of Tariff and ActiveEnergy.

    `energy`: int 4byte signed; None if invalid
    `tariff`: int 0..254; None if invalid
    """

    energy: int | None = None
    tariff: int | None = None

    @classmethod
    def from_dict(cls, data: Mapping[str, int]) -> TariffActiveEnergy:
        """Init from a dictionary."""
        try:
            energy = data.get("energy")
            tariff = data.get("tariff")
        except AttributeError as err:
            raise ValueError(f"Invalid value for TariffActiveEnergy: {err}") from err
        if energy is not None:
            try:
                energy = int(energy)
            except ValueError as err:
                raise ValueError(f"Invalid value for energy: {err}") from err
        if tariff is not None:
            try:
                tariff = int(tariff)
            except ValueError as err:
                raise ValueError(f"Invalid value for tariff: {err}") from err
        return cls(energy=energy, tariff=tariff)

    def as_dict(self) -> dict[str, int | None]:
        """Create a JSON serializable dictionary."""
        return {
            "energy": self.energy,
            "tariff": self.tariff,
        }


class DPTTariffActiveEnergy(DPTComplex[TariffActiveEnergy]):
    """Abstraction for KNX 6 octet Tariff and ActiveEnergy (DPT 235.001)."""

    data_type = TariffActiveEnergy
    payload_type = DPTArray
    payload_length = 6
    dpt_main_number = 235
    dpt_sub_number = 1
    value_type = "tariff_active_energy"

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> TariffActiveEnergy:
        """Parse/deserialize from KNX/IP raw data."""
        raw = cls.validate_payload(payload)
        energy_valid = (raw[5] >> 1 & 0b1) == 0
        tariff_valid = (raw[5] & 0b1) == 0
        return TariffActiveEnergy(
            energy=(
                DPTActiveEnergy.from_knx(DPTArray(raw[:4])) if energy_valid else None
            ),
            tariff=DPTTariff.from_knx(DPTArray([raw[4]])) if tariff_valid else None,
        )

    @classmethod
    def _to_knx(cls, value: TariffActiveEnergy) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        energy = (
            DPTActiveEnergy.to_knx(value.energy).value
            if value.energy is not None
            else [0, 0, 0, 0]
        )
        tariff = (
            DPTTariff.to_knx(value.tariff).value if value.tariff is not None else [0]
        )
        return DPTArray(
            (
                *energy,
                *tariff,
                (value.energy is None) << 1 | (value.tariff is None),
            )
        )



"""Unit test for KNX/IP RoutingLostMessage objects."""

import pytest

from xknx.exceptions import CouldNotParseKNXIP
from xknx.knxip import KNXIPFrame, RoutingLostMessage


class TestKNXIPRoutingLostMessage:
    """Test class for KNX/IP RoutingLostMessage objects."""

    def test_routing_lost_message(self) -> None:
        """Test parsing and streaming RoutingLostMessage KNX/IP packet."""
        raw = bytes((0x06, 0x10, 0x05, 0x31, 0x00, 0x0A, 0x04, 0x00, 0x00, 0x05))
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, RoutingLostMessage)
        assert knxipframe.body.device_state == 0
        assert knxipframe.body.lost_messages == 5

        routing_lost_message = RoutingLostMessage(lost_messages=5)
        knxipframe2 = KNXIPFrame.init_from_body(routing_lost_message)

        assert knxipframe2.to_knx() == raw

    def test_from_knx_wrong_lost_message_information(self) -> None:
        """Test parsing and streaming wrong RoutingLostMessage (wrong length byte)."""
        raw = bytes((0x06, 0x10, 0x05, 0x31, 0x00, 0x0A, 0x06, 0x00, 0x00, 0x05))
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_from_knx_wrong_lost_message_information2(self) -> None:
        """Test parsing and streaming wrong RoutingLostMessage (wrong length)."""
        raw = bytes((0x06, 0x10, 0x05, 0x31, 0x00, 0x0A, 0x04, 0x00, 0x00))
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)



"""Unit test for KNX/IP Interface."""

from collections.abc import AsyncGenerator
from pathlib import Path
import threading
from typing import Any
from unittest.mock import DEFAULT, Mock, patch

import pytest

from xknx import XKNX
from xknx.exceptions.exception import CommunicationError, InvalidSecureConfiguration
from xknx.io import (
    ConnectionConfig,
    ConnectionType,
    GatewayDescriptor,
    SecureConfig,
    knx_interface_factory,
)
from xknx.io.routing import Routing, SecureGroup, SecureRouting
from xknx.io.tunnel import SecureTunnel, TCPTunnel, UDPTunnel
from xknx.knxip.dib import TunnelingSlotStatus
from xknx.secure import load_keyring
from xknx.telegram import IndividualAddress


class TestKNXIPInterface:
    """Test class for KNX interface objects."""

    knxkeys_file = Path(__file__).parent / "resources/testcase.knxkeys"

    def setup_method(self) -> None:
        """Set up test class."""
        # pylint: disable=attribute-defined-outside-init
        self.xknx = XKNX()

    async def test_start_automatic_connection(self) -> None:
        """Test starting automatic connection."""
        connection_config = ConnectionConfig()
        assert connection_config.connection_type == ConnectionType.AUTOMATIC
        interface = knx_interface_factory(self.xknx, connection_config)
        with patch("xknx.io.KNXIPInterface._start_automatic") as start_automatic_mock:
            await interface.start()
            start_automatic_mock.assert_called_once_with(local_ip=None, keyring=None)
            assert threading.active_count() == 1

        async def gateway_generator_mock(_: Any) -> AsyncGenerator[GatewayDescriptor]:
            secure_interface = GatewayDescriptor(
                ip_addr="10.1.2.3",
                port=3671,
                supports_tunnelling_tcp=True,
                supports_secure=True,
            )
            secure_interface.tunnelling_requires_secure = True
            yield secure_interface
            yield GatewayDescriptor(
                ip_addr="10.1.2.3", port=3671, supports_tunnelling_tcp=True
            )
            yield GatewayDescriptor(
                ip_addr="10.1.2.3", port=3671, supports_tunnelling=True
            )
            yield GatewayDescriptor(
                ip_addr="10.1.2.3", port=3671, supports_routing=True
            )

        with (
            patch(
                "xknx.io.knxip_interface.GatewayScanner.async_scan",
                new=gateway_generator_mock,
            ),
            patch(
                "xknx.io.KNXIPInterface._start_secure_tunnelling_tcp",
                side_effect=InvalidSecureConfiguration("Error"),
            ) as start_secure_tunnelling_tcp,
            patch(
                "xknx.io.KNXIPInterface._start_tunnelling_tcp",
                side_effect=CommunicationError("Error"),
            ) as start_tunnelling_tcp_mock,
            patch(
                "xknx.io.KNXIPInterface._start_tunnelling_udp",
                side_effect=CommunicationError("Error"),
            ) as start_tunnelling_udp_mock,
            patch(
                "xknx.io.KNXIPInterface._start_routing",
                side_effect=CommunicationError("Error"),
            ) as start_routing_mock,
        ):
            with pytest.raises(CommunicationError):
                await interface.start()
            start_secure_tunnelling_tcp.assert_called_once()
            start_tunnelling_tcp_mock.assert_called_once()
            start_tunnelling_udp_mock.assert_called_once()
            start_routing_mock.assert_called_once()

    async def test_start_automatic_with_keyring(self) -> None:
        """Test starting with automatic mode and keyring."""
        connection_config = ConnectionConfig(
            secure_config=SecureConfig(
                knxkeys_file_path=self.knxkeys_file,
                knxkeys_password="password",
            )
        )
        assert connection_config.connection_type == ConnectionType.AUTOMATIC
        interface = knx_interface_factory(self.xknx, connection_config)

        # in the test keyfile the only host is 1.0.0 - others shall be skipped
        async def gateway_generator_mock(_: Any) -> AsyncGenerator[GatewayDescriptor]:
            yield GatewayDescriptor(
                ip_addr="10.1.5.5",
                port=3671,
                supports_tunnelling_tcp=True,
                individual_address=IndividualAddress("5.0.0"),
            )
            yield GatewayDescriptor(
                ip_addr="10.1.0.0",
                port=3671,
                supports_tunnelling_tcp=True,
                individual_address=IndividualAddress("1.0.0"),
            )

        with (
            patch(
                "xknx.io.knxip_interface.GatewayScanner.async_scan",
                new=gateway_generator_mock,
            ),
            patch(
                "xknx.io.KNXIPInterface._start_tunnelling_tcp",
            ) as start_tunnelling_tcp_mock,
        ):
            await interface.start()
            start_tunnelling_tcp_mock.assert_called_once_with(
                gateway_ip="10.1.0.0",
                gateway_port=3671,
            )

    async def test_start_automatic_with_keyring_and_ia(self) -> None:
        """Test starting with automatic mode and keyring and individual address."""
        connection_config = ConnectionConfig(
            individual_address=IndividualAddress("1.0.12"),
            secure_config=SecureConfig(
                knxkeys_file_path=self.knxkeys_file,
                knxkeys_password="password",
            ),
        )
        assert connection_config.connection_type == ConnectionType.AUTOMATIC
        interface = knx_interface_factory(self.xknx, connection_config)

        # in the test keyfile the only host is 1.0.0 - others shall be skipped
        async def gateway_generator_mock(_: Any) -> AsyncGenerator[GatewayDescriptor]:
            yield GatewayDescriptor(
                ip_addr="10.1.5.5",
                port=3671,
                supports_tunnelling_tcp=True,
                individual_address=IndividualAddress("5.0.0"),
            )
            yield GatewayDescriptor(
                ip_addr="10.1.0.0",
                port=3671,
                supports_tunnelling_tcp=True,
                individual_address=IndividualAddress("1.0.0"),
            )

        with (
            patch(
                "xknx.io.knxip_interface.GatewayScanner.async_scan",
                new=gateway_generator_mock,
            ),
            patch(
                "xknx.io.KNXIPInterface._start_tunnelling_tcp",
            ) as start_tunnelling_tcp_mock,
        ):
            await interface.start()
            start_tunnelling_tcp_mock.assert_called_once_with(
                gateway_ip="10.1.0.0",
                gateway_port=3671,
            )

        # IA not listed in keyring
        invalid_config = ConnectionConfig(
            individual_address=IndividualAddress("5.5.5"),
            secure_config=SecureConfig(
                knxkeys_file_path=self.knxkeys_file,
                knxkeys_password="password",
            ),
        )
        assert invalid_config.connection_type == ConnectionType.AUTOMATIC
        interface = knx_interface_factory(self.xknx, invalid_config)
        with (
            patch(
                "xknx.io.knxip_interface.GatewayScanner.async_scan",
                new=gateway_generator_mock,
            ),
            pytest.raises(InvalidSecureConfiguration),
        ):
            await interface.start()

    async def test_start_udp_tunnel_connection(self) -> None:
        """Test starting UDP tunnel connection."""
        # without gateway_ip automatic is called
        gateway_ip = "127.0.0.2"
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.TUNNELING,
            gateway_ip=gateway_ip,
        )
        with patch(
            "xknx.io.KNXIPInterface._start_tunnelling_udp"
        ) as start_tunnelling_udp:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            start_tunnelling_udp.assert_called_once_with(
                gateway_ip=gateway_ip,
                gateway_port=3671,
                local_ip=None,
            )
        with patch("xknx.io.tunnel.UDPTunnel.connect") as connect_udp:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            assert isinstance(interface._interface, UDPTunnel)
            assert interface._interface.local_ip == "127.0.0.1"
            assert interface._interface.local_port == 0
            assert interface._interface.gateway_ip == gateway_ip
            assert interface._interface.gateway_port == 3671
            assert interface._interface.auto_reconnect
            assert interface._interface.auto_reconnect_wait == 3
            assert interface._interface.route_back is False
            assert (  # pylint: disable=comparison-with-callable
                interface._interface.cemi_received_callback == interface.cemi_received
            )
            connect_udp.assert_called_once_with()

    async def test_start_tcp_tunnel_connection(self) -> None:
        """Test starting TCP tunnel connection."""
        # without gateway_ip automatic is called
        gateway_ip = "127.0.0.2"
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.TUNNELING_TCP, gateway_ip=gateway_ip
        )
        with patch(
            "xknx.io.KNXIPInterface._start_tunnelling_tcp"
        ) as start_tunnelling_tcp:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            start_tunnelling_tcp.assert_called_once_with(
                gateway_ip=gateway_ip,
                gateway_port=3671,
            )
        with patch("xknx.io.tunnel.TCPTunnel.connect") as connect_tcp:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            assert isinstance(interface._interface, TCPTunnel)
            assert interface._interface.gateway_ip == gateway_ip
            assert interface._interface.gateway_port == 3671
            assert interface._interface.auto_reconnect
            assert interface._interface.auto_reconnect_wait == 3
            assert (  # pylint: disable=comparison-with-callable
                interface._interface.cemi_received_callback == interface.cemi_received
            )
            connect_tcp.assert_called_once_with()

    async def test_start_tcp_tunnel_connection_with_ia(self) -> None:
        """Test starting TCP tunnel connection requesting specific tunnel."""
        # without gateway_ip automatic is called
        gateway_ip = "127.0.0.2"
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.TUNNELING_TCP,
            gateway_ip=gateway_ip,
            individual_address="1.1.1",
        )
        with patch(
            "xknx.io.KNXIPInterface._start_tunnelling_tcp"
        ) as start_tunnelling_tcp:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            start_tunnelling_tcp.assert_called_once_with(
                gateway_ip=gateway_ip,
                gateway_port=3671,
            )
        with patch("xknx.io.tunnel.TCPTunnel.connect") as connect_tcp:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            assert isinstance(interface._interface, TCPTunnel)
            assert interface._interface.gateway_ip == gateway_ip
            assert interface._interface.gateway_port == 3671
            assert interface._interface._requested_address == IndividualAddress("1.1.1")
            assert interface._interface.auto_reconnect
            assert interface._interface.auto_reconnect_wait == 3
            assert (  # pylint: disable=comparison-with-callable
                interface._interface.cemi_received_callback == interface.cemi_received
            )
            connect_tcp.assert_called_once_with()

    async def test_start_routing_connection(self) -> None:
        """Test starting routing connection."""
        local_ip = "127.0.0.1"
        # set local_ip to avoid gateway scanner
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.ROUTING, local_ip=local_ip
        )
        with patch("xknx.io.KNXIPInterface._start_routing") as start_routing:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            start_routing.assert_called_once_with(local_ip=local_ip)

        with patch("xknx.io.routing.Routing.connect") as connect_routing:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            assert isinstance(interface._interface, Routing)
            assert interface._interface.local_ip == local_ip
            assert interface._interface.multicast_group == "224.0.23.12"
            assert interface._interface.multicast_port == 3671
            assert (  # pylint: disable=comparison-with-callable
                interface._interface.cemi_received_callback == interface.cemi_received
            )
            connect_routing.assert_called_once_with()

    async def test_threaded_connection(self) -> None:
        """Test starting threaded connection."""
        # pylint: disable=attribute-defined-outside-init
        self.main_thread = threading.get_ident()

        def assert_thread(*args: Any, **kwargs: dict[str, Any]) -> None:
            """Test threaded connection."""
            assert self.main_thread != threading.get_ident()

        connection_config = ConnectionConfig(threaded=True)
        assert connection_config.connection_type == ConnectionType.AUTOMATIC
        with patch(
            "xknx.io.KNXIPInterface._start_automatic", side_effect=assert_thread
        ) as start_automatic_mock:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            start_automatic_mock.assert_called_once_with(local_ip=None, keyring=None)

            with pytest.raises(CommunicationError):
                # interface can only be started once
                await interface.start()

    async def test_threaded_send_cemi(self) -> None:
        """Test sending cemi with threaded connection."""
        # pylint: disable=attribute-defined-outside-init
        self.main_thread = threading.get_ident()

        def assert_thread(*args: Any, **kwargs: dict[str, Any]) -> Any:
            """Test threaded connection."""
            assert self.main_thread != threading.get_ident()
            return DEFAULT  # to not disable `return_value` of send_cemi_mock

        local_ip = "127.0.0.1"
        # set local_ip to avoid gateway scanner; use routing as it is the simplest mode
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.ROUTING, local_ip=local_ip, threaded=True
        )
        cemi_mock = Mock()
        with (
            patch(
                "xknx.io.routing.Routing.connect", side_effect=assert_thread
            ) as connect_routing_mock,
            patch(
                "xknx.io.routing.Routing.send_cemi",
                side_effect=assert_thread,
                return_value="test",
            ) as send_cemi_mock,
            patch(
                "xknx.io.routing.Routing.disconnect", side_effect=assert_thread
            ) as disconnect_routing_mock,
        ):
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            connect_routing_mock.assert_called_once_with()
            assert await interface.send_cemi(cemi_mock) == "test"
            send_cemi_mock.assert_called_once_with(cemi_mock)
            await interface.stop()
            disconnect_routing_mock.assert_called_once_with()
            assert interface._interface is None

    async def test_threaded_connection_unsuccessful_start(self) -> None:
        """Test cleanup when unsuccessful initial connection."""
        connection_config = ConnectionConfig(threaded=True)
        assert connection_config.connection_type == ConnectionType.AUTOMATIC
        with patch(
            "xknx.io.KNXIPInterface._start", side_effect=CommunicationError("Error")
        ):
            interface = knx_interface_factory(self.xknx, connection_config)
            with pytest.raises(CommunicationError):
                await interface.start()
            # thread and loop are cleaned up after unsuccessful start
            assert interface._connection_thread is None
            assert interface._thread_loop is None

    async def test_start_secure_connection_knx_keys_user_id(self) -> None:
        """Test starting a secure connection from a knxkeys file with user_id."""
        gateway_ip = "192.168.1.1"
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.TUNNELING_TCP_SECURE,
            gateway_ip=gateway_ip,
            secure_config=SecureConfig(
                user_id=3,
                knxkeys_file_path=self.knxkeys_file,
                knxkeys_password="password",
            ),
        )
        gateway_description = GatewayDescriptor(
            ip_addr=gateway_ip,
            port=3671,
            supports_tunnelling_tcp=True,
            supports_secure=True,
            individual_address=IndividualAddress("1.0.0"),
        )
        gateway_description.tunnelling_requires_secure = True
        with (
            patch(
                "xknx.io.knxip_interface.request_description",
                return_value=gateway_description,
            ),
            patch("xknx.io.tunnel.SecureTunnel.connect") as connect_secure,
        ):
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            assert isinstance(interface._interface, SecureTunnel)
            assert interface._interface.gateway_ip == gateway_ip
            assert interface._interface.gateway_port == 3671
            assert interface._interface.auto_reconnect is True
            assert interface._interface.auto_reconnect_wait == 3
            assert interface._interface._user_id == 3
            assert interface._interface._user_password == "user1"
            assert (
                interface._interface._device_authentication_password
                == "authenticationcode"
            )
            assert (  # pylint: disable=comparison-with-callable
                interface._interface.cemi_received_callback == interface.cemi_received
            )
            connect_secure.assert_called_once_with()

    async def test_start_secure_connection_knx_keys_ia(self) -> None:
        """Test starting a secure connection from a knxkeys file with individual address."""
        gateway_ip = "192.168.1.1"
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.TUNNELING_TCP_SECURE,
            gateway_ip=gateway_ip,
            individual_address="1.0.12",
            secure_config=SecureConfig(
                knxkeys_file_path=self.knxkeys_file, knxkeys_password="password"
            ),
        )
        # result of request_description is currently not used when IA is defined
        with (
            patch(
                "xknx.io.knxip_interface.request_description",
            ),
            patch("xknx.io.tunnel.SecureTunnel.connect") as connect_secure,
        ):
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            assert isinstance(interface._interface, SecureTunnel)
            assert interface._interface.gateway_ip == gateway_ip
            assert interface._interface.gateway_port == 3671
            assert interface._interface.auto_reconnect is True
            assert interface._interface.auto_reconnect_wait == 3
            assert interface._interface._user_id == 5
            assert interface._interface._user_password == "user3"
            assert (
                interface._interface._device_authentication_password
                == "authenticationcode"
            )
            assert (  # pylint: disable=comparison-with-callable
                interface._interface.cemi_received_callback == interface.cemi_received
            )
            connect_secure.assert_called_once_with()

    async def test_start_secure_connection_knx_keys_first_interface(self) -> None:
        """Test starting a secure connection from a knxkeys file."""
        gateway_ip = "192.168.1.1"
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.TUNNELING_TCP_SECURE,
            gateway_ip=gateway_ip,
            secure_config=SecureConfig(
                knxkeys_file_path=self.knxkeys_file, knxkeys_password="password"
            ),
        )
        gateway_description = GatewayDescriptor(
            ip_addr=gateway_ip,
            port=3671,
            supports_tunnelling_tcp=True,
            supports_secure=True,
            individual_address=IndividualAddress("1.0.0"),
        )
        gateway_description.tunnelling_requires_secure = True
        gateway_description.tunnelling_slots = {
            IndividualAddress("1.0.1"): TunnelingSlotStatus(
                usable=True, authorized=False, free=False
            ),
            IndividualAddress("1.0.11"): TunnelingSlotStatus(
                usable=True, authorized=False, free=True
            ),
            IndividualAddress("1.0.12"): TunnelingSlotStatus(
                usable=True, authorized=False, free=True
            ),
            IndividualAddress("1.0.13"): TunnelingSlotStatus(
                usable=True, authorized=False, free=True
            ),
        }
        with (
            patch(
                "xknx.io.knxip_interface.request_description",
                return_value=gateway_description,
            ),
            patch("xknx.io.tunnel.SecureTunnel.connect") as connect_secure,
        ):
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            assert isinstance(interface._interface, SecureTunnel)
            assert interface._interface.gateway_ip == gateway_ip
            assert interface._interface.gateway_port == 3671
            assert interface._interface.auto_reconnect is True
            assert interface._interface.auto_reconnect_wait == 3
            assert interface._interface._user_id == 4
            assert interface._interface._user_password == "user2"
            assert (
                interface._interface._device_authentication_password
                == "authenticationcode"
            )
            assert (  # pylint: disable=comparison-with-callable
                interface._interface.cemi_received_callback == interface.cemi_received
            )
            connect_secure.assert_called_once_with()

    async def test_start_secure_with_manual_passwords(self) -> None:
        """Test starting a secure connection from manual passwords."""
        gateway_ip = "192.168.1.1"
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.TUNNELING_TCP_SECURE,
            gateway_ip=gateway_ip,
            secure_config=SecureConfig(
                user_id=3,
                device_authentication_password="authenticationcode",
                user_password="user1",
            ),
        )
        with patch("xknx.io.tunnel.SecureTunnel.connect") as connect_secure:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            assert isinstance(interface._interface, SecureTunnel)
            assert interface._interface.gateway_ip == gateway_ip
            assert interface._interface.gateway_port == 3671
            assert interface._interface.auto_reconnect is True
            assert interface._interface.auto_reconnect_wait == 3
            assert interface._interface._user_id == 3
            assert interface._interface._user_password == "user1"
            assert (
                interface._interface._device_authentication_password
                == "authenticationcode"
            )
            assert (  # pylint: disable=comparison-with-callable
                interface._interface.cemi_received_callback == interface.cemi_received
            )
            connect_secure.assert_called_once_with()

    @pytest.mark.parametrize(
        "connection_config",
        [
            ConnectionConfig(  # invalid user_id
                connection_type=ConnectionType.TUNNELING_TCP_SECURE,
                gateway_ip="192.168.1.1",
                secure_config=SecureConfig(
                    user_id=12,
                    knxkeys_file_path=knxkeys_file,
                    knxkeys_password="password",
                ),
            ),
            ConnectionConfig(  # invalid IA
                connection_type=ConnectionType.TUNNELING_TCP_SECURE,
                gateway_ip="192.168.1.1",
                individual_address="9.9.9",
                secure_config=SecureConfig(
                    knxkeys_file_path=knxkeys_file,
                    knxkeys_password="password",
                ),
            ),
            ConnectionConfig(  # no secure_config
                connection_type=ConnectionType.TUNNELING_TCP_SECURE,
                gateway_ip="192.168.1.1",
                individual_address="9.9.9",
            ),
            ConnectionConfig(  # no password
                connection_type=ConnectionType.TUNNELING_TCP_SECURE,
                gateway_ip="192.168.1.1",
                secure_config=SecureConfig(
                    user_id=3,
                    knxkeys_file_path=knxkeys_file,
                ),
            ),
        ],
    )
    async def test_invalid_secure_error(
        self, connection_config: ConnectionConfig
    ) -> None:
        """Test ip secure invalid configurations."""
        gateway_ip = "192.168.1.1"
        gateway_description = GatewayDescriptor(
            ip_addr=gateway_ip,
            port=3671,
            supports_tunnelling_tcp=True,
            supports_secure=True,
            individual_address=IndividualAddress("1.0.0"),
        )
        gateway_description.tunnelling_requires_secure = True
        with (
            patch(
                "xknx.io.knxip_interface.request_description",
                return_value=gateway_description,
            ),
            pytest.raises(InvalidSecureConfiguration),
        ):
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()

    async def test_invalid_user_password(self) -> None:
        """Test ip secure."""
        gateway_ip = "192.168.1.1"
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.TUNNELING_TCP_SECURE,
            gateway_ip=gateway_ip,
            secure_config=SecureConfig(
                user_id=1,
            ),
        )
        with pytest.raises(InvalidSecureConfiguration):
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()

    async def test_start_secure_routing_knx_keys(self) -> None:
        """Test starting a secure routing connection from a knxkeys file."""
        backbone_key = bytes.fromhex("cf89fd0f18f4889783c7ef44ee1f5e14")
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.ROUTING_SECURE,
            secure_config=SecureConfig(
                knxkeys_file_path=self.knxkeys_file, knxkeys_password="password"
            ),
        )
        with patch("xknx.io.routing.SecureRouting.connect") as connect_secure:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            assert isinstance(interface._interface, SecureRouting)
            assert interface._interface.backbone_key == backbone_key
            assert interface._interface.latency_ms == 1000
            assert isinstance(interface._interface.transport, SecureGroup)
            assert interface._interface.transport.remote_addr == (
                "224.0.23.12",
                3671,
            )
            assert (  # pylint: disable=comparison-with-callable
                interface._interface.cemi_received_callback == interface.cemi_received
            )
            connect_secure.assert_called_once_with()

    async def test_start_secure_routing_manual(self) -> None:
        """Test starting a secure routing connection from a knxkeys file."""
        backbone_key_str = "cf89fd0f18f4889783c7ef44ee1f5e14"
        backbone_key = bytes.fromhex(backbone_key_str)
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.ROUTING_SECURE,
            secure_config=SecureConfig(backbone_key=backbone_key_str, latency_ms=2000),
        )
        with patch("xknx.io.routing.SecureRouting.connect") as connect_secure:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            assert isinstance(interface._interface, SecureRouting)
            assert interface._interface.backbone_key == backbone_key
            assert interface._interface.latency_ms == 2000
            assert isinstance(interface._interface.transport, SecureGroup)
            assert interface._interface.transport.remote_addr == (
                "224.0.23.12",
                3671,
            )
            assert (  # pylint: disable=comparison-with-callable
                interface._interface.cemi_received_callback == interface.cemi_received
            )
            connect_secure.assert_called_once_with()

    async def test_start_secure_routing_explicit_keyring(self) -> None:
        """Test starting a secure routing connection with a pre-loaded keyring."""
        backbone_key = bytes.fromhex("cf89fd0f18f4889783c7ef44ee1f5e14")
        explicit_keyring = await load_keyring(
            path=self.knxkeys_file, password="password"
        )
        connection_config = ConnectionConfig(
            connection_type=ConnectionType.ROUTING_SECURE,
            secure_config=SecureConfig(keyring=explicit_keyring),
        )
        with patch("xknx.io.routing.SecureRouting.connect") as connect_secure:
            interface = knx_interface_factory(self.xknx, connection_config)
            await interface.start()
            assert isinstance(interface._interface, SecureRouting)
            assert interface._interface.backbone_key == backbone_key
            assert interface._interface.latency_ms == 1000
            assert isinstance(interface._interface.transport, SecureGroup)
            assert interface._interface.transport.remote_addr == (
                "224.0.23.12",
                3671,
            )
            assert (  # pylint: disable=comparison-with-callable
                interface._interface.cemi_received_callback == interface.cemi_received
            )
            connect_secure.assert_called_once_with()



"""
Abstraction for handling KNXnet/IP routing.

Routing uses UDP Multicast to send and receive KNXnet/IP messages.
"""

from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
import logging
import random
from typing import TYPE_CHECKING, Final

from xknx.cemi import CEMIFrame, CEMIMessageCode
from xknx.core import XknxConnectionState, XknxConnectionType
from xknx.exceptions import CommunicationError
from xknx.knxip import (
    HPAI,
    KNXIPFrame,
    KNXIPServiceType,
    RoutingBusy,
    RoutingIndication,
    RoutingLostMessage,
)
from xknx.telegram import IndividualAddress

from .const import DEFAULT_INDIVIDUAL_ADDRESS, DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT
from .interface import CEMIBytesCallbackType, Interface
from .ip_secure import SecureGroup
from .transport import KNXIPTransport, UDPTransport

if TYPE_CHECKING:
    from xknx.xknx import XKNX

logger = logging.getLogger("xknx.log")

BUSY_DECREMENT_TIME: Final = 0.005  # 5 ms
BUSY_INCREMENT_COOLDOWN: Final = 0.01  # 10 ms
BUSY_RANDOM_TIME_FACTOR: Final = 0.05  # 50 ms
BUSY_SLOWDURATION_TIME_FACTOR: Final = 0.1  # 100 ms
ROUTING_INDICATION_WAIT_TIME: Final = 0.02  # 20 ms

DEFAULT_LATENCY_TOLERANCE_MS: Final = 1000


class _RoutingFlowControl:
    """
    Class for handling KNXnet/IP routing flow control.

    See KNX Specifications 3.8.5 Routing §2.3.5 Flow control handling
    """

    __slots__ = (
        "_last_busy_frame_time",
        "_last_sent_routing_indication_time",
        "_loop",
        "_ready",
        "_received_busy_frames",
        "_timer_task",
        "_wait_start_time",
        "_wait_time_ms",
    )

    def __init__(self) -> None:
        self._last_busy_frame_time: float = 0.0
        self._last_sent_routing_indication_time: float = 0.0
        self._loop = asyncio.get_running_loop()
        self._ready = asyncio.Event()
        self._ready.set()
        self._received_busy_frames: int = 0
        self._timer_task: asyncio.Task[None] | None = None
        self._wait_start_time: float | None = None
        self._wait_time_ms: int = 0

    def cancel(self) -> None:
        """Cancel internal tasks."""
        if self._timer_task:
            self._timer_task.cancel()

    @asynccontextmanager
    async def throttle(self) -> AsyncIterator[None]:
        """Context manager to wait for ready state and throttle outgoing frames."""
        # limit RoutingIndication transmission rate according to
        # KNX Specifications 3.2.6 Communication Medium KNX IP §2.1
        # simplified version - pause 20 ms after transmit a RoutingIndication
        elapsed = self._loop.time() - self._last_sent_routing_indication_time
        if elapsed < ROUTING_INDICATION_WAIT_TIME:
            await asyncio.sleep(ROUTING_INDICATION_WAIT_TIME - elapsed)

        await self._ready.wait()
        yield
        self._last_sent_routing_indication_time = self._loop.time()

    def handle_routing_busy(self, routing_busy: RoutingBusy) -> None:
        """Handle incoming RoutingBusy."""
        self._ready.clear()
        now = self._loop.time()
        previous_busy_frame_time = self._last_busy_frame_time
        self._last_busy_frame_time = now
        if self._wait_start_time is None:
            logger.info(
                "RoutingBusy received: %s",
                routing_busy,
            )
        else:
            # only apply if we have already received a RoutingBusy frame and are still pausing
            if (now - previous_busy_frame_time) > BUSY_INCREMENT_COOLDOWN:
                self._received_busy_frames += 1
            logger.debug(
                "RoutingBusy received: %s - %s ms since previous - number %s in moving time window",
                routing_busy,
                round((now - previous_busy_frame_time) * 1000),
                self._received_busy_frames,
            )
            # discard frame if wait time is lower than remaining time
            elapsed_ms = (now - self._wait_start_time) * 1000
            remaining_ms = self._wait_time_ms - elapsed_ms
            if remaining_ms >= routing_busy.wait_time:
                return
        self._wait_time_ms = routing_busy.wait_time
        self._wait_start_time = now

        if self._timer_task:
            self._timer_task.cancel()
        self._timer_task = asyncio.create_task(self._resume_sending())

    async def _resume_sending(self) -> None:
        """Reset ready flag after wait_time_ms and fade out slowduration."""
        random_wait_extension = (
            random.random() * self._received_busy_frames * BUSY_RANDOM_TIME_FACTOR
        )
        slowduration = self._received_busy_frames * BUSY_SLOWDURATION_TIME_FACTOR
        await asyncio.sleep(self._wait_time_ms / 1000 + random_wait_extension)

        self._ready.set()
        self._wait_start_time = None
        await asyncio.sleep(slowduration)
        while self._received_busy_frames > 0:
            await asyncio.sleep(BUSY_DECREMENT_TIME)
            self._received_busy_frames -= 1


class Routing(Interface):
    """Class for handling KNXnet/IP multicast communication."""

    __slots__ = (
        "_flow_control",
        "cemi_received_callback",
        "individual_address",
        "local_ip",
        "multicast_group",
        "multicast_port",
        "xknx",
    )

    connection_type = XknxConnectionType.ROUTING
    transport: UDPTransport

    def __init__(
        self,
        xknx: XKNX,
        individual_address: IndividualAddress | None,
        cemi_received_callback: CEMIBytesCallbackType,
        local_ip: str,
        multicast_group: str = DEFAULT_MCAST_GRP,
        multicast_port: int = DEFAULT_MCAST_PORT,
    ) -> None:
        """Initialize Routing class."""
        self.xknx = xknx
        self.individual_address = individual_address or DEFAULT_INDIVIDUAL_ADDRESS
        self.cemi_received_callback = cemi_received_callback
        self.local_ip = local_ip
        self.multicast_group = multicast_group
        self.multicast_port = multicast_port

        self._init_transport()
        self.transport.register_callback(
            self._handle_frame,
            [
                KNXIPServiceType.ROUTING_INDICATION,
                KNXIPServiceType.ROUTING_BUSY,
                KNXIPServiceType.ROUTING_LOST_MESSAGE,
            ],
        )
        self._flow_control = _RoutingFlowControl()

    def _init_transport(self) -> None:
        """Initialize transport."""
        self.transport = UDPTransport(
            local_addr=(self.local_ip, 0),
            remote_addr=(self.multicast_group, self.multicast_port),
            multicast=True,
        )

    ####################
    #
    # CONNECT DISCONNECT
    #
    ####################

    async def connect(self) -> None:
        """Start routing."""
        self.xknx.current_address = self.individual_address
        self.xknx.connection_manager.connection_state_changed(
            XknxConnectionState.CONNECTING, self.connection_type
        )
        try:
            await self.transport.connect()
        except OSError as ex:
            logger.debug(
                "Could not establish connection to KNXnet/IP network. %s: %s",
                type(ex).__name__,
                ex,
            )
            self.xknx.connection_manager.connection_state_changed(
                XknxConnectionState.DISCONNECTED
            )
            # close udp transport to prevent open file descriptors
            self.transport.stop()
            raise CommunicationError("Routing could not be started") from ex
        self.xknx.connection_manager.connection_state_changed(
            XknxConnectionState.CONNECTED, self.connection_type
        )

    async def disconnect(self) -> None:
        """Stop routing."""
        self.transport.stop()
        self.xknx.connection_manager.connection_state_changed(
            XknxConnectionState.DISCONNECTED
        )
        self._flow_control.cancel()

    ##################
    #
    # OUTGOING FRAMES
    #
    ##################

    async def send_cemi(self, cemi: CEMIFrame) -> None:
        """Send CEMIFrame to the network."""
        # send L_DATA_IND to network, create L_DATA_CON locally for routing
        cemi.code = CEMIMessageCode.L_DATA_IND
        routing_indication = RoutingIndication(raw_cemi=cemi.to_knx())

        async with self._flow_control.throttle():
            self._send_knxipframe(KNXIPFrame.init_from_body(routing_indication))

        cemi.code = CEMIMessageCode.L_DATA_CON
        self.cemi_received_callback(cemi.to_knx())

    def _send_knxipframe(self, knxipframe: KNXIPFrame) -> None:
        """Send KNXIPFrame to connected routing device."""
        self.transport.send(knxipframe)

    ##################
    #
    # INCOMING FRAMES
    #
    ##################

    def _handle_frame(
        self, knxipframe: KNXIPFrame, source: HPAI, _: KNXIPTransport
    ) -> None:
        """Handle incoming KNXIPFrames. Callback from internal transport."""
        if isinstance(knxipframe.body, RoutingIndication):
            self._handle_routing_indication(knxipframe.body)
        elif isinstance(knxipframe.body, RoutingBusy):
            self._flow_control.handle_routing_busy(knxipframe.body)
        elif isinstance(knxipframe.body, RoutingLostMessage):
            logger.warning(
                "RoutingLostMessage received from %s - %s lost messages.",
                source.ip_addr,
                knxipframe.body.lost_messages,
            )
        else:
            logger.warning("Service not implemented: %s", knxipframe)

    def _handle_routing_indication(self, routing_indication: RoutingIndication) -> None:
        """Handle incoming RoutingIndication."""
        self.cemi_received_callback(routing_indication.raw_cemi)


class SecureRouting(Routing):
    """Class for handling KNXnet/IP secure multicast communication."""

    __slots__ = (
        "backbone_key",
        "latency_ms",
    )

    connection_type = XknxConnectionType.ROUTING_SECURE
    transport: SecureGroup

    def __init__(
        self,
        xknx: XKNX,
        individual_address: IndividualAddress | None,
        cemi_received_callback: CEMIBytesCallbackType,
        local_ip: str,
        backbone_key: bytes,
        latency_ms: int | None = None,
        multicast_group: str = DEFAULT_MCAST_GRP,
        multicast_port: int = DEFAULT_MCAST_PORT,
    ) -> None:
        """Initialize SecureRouting class."""
        self.backbone_key = backbone_key
        self.latency_ms = latency_ms or DEFAULT_LATENCY_TOLERANCE_MS
        super().__init__(
            xknx,
            individual_address=individual_address,
            cemi_received_callback=cemi_received_callback,
            local_ip=local_ip,
            multicast_group=multicast_group,
            multicast_port=multicast_port,
        )

    def _init_transport(self) -> None:
        """Initialize transport."""
        self.transport = SecureGroup(
            local_addr=(self.local_ip, 0),
            remote_addr=(self.multicast_group, self.multicast_port),
            backbone_key=self.backbone_key,
            latency_ms=self.latency_ms,
        )



"""Example for running a callback when a devices state changed."""

import asyncio

from xknx import XKNX
from xknx.devices import Light
from xknx.io import ConnectionConfig, ConnectionType


def light_callback(light: Light) -> None:
    """Run callback when the light changed any of its state."""
    print(f"{light.name} - {light.state}")


async def main() -> None:
    """Connect to KNX/IP bus, turn on Light device and wait."""
    xknx = XKNX(
        connection_config=ConnectionConfig(
            gateway_ip="10.1.0.41",
            connection_type=ConnectionType.TUNNELING_TCP,
        ),
    )
    await xknx.start()
    light = Light(
        xknx,
        name="TestLight",
        group_address_switch="1/1/45",
        group_address_switch_state="1/0/45",
        device_updated_cb=light_callback,
    )
    xknx.devices.async_add(light)
    # turn on light and listen for 10 seconds for changes
    await light.set_on()
    await asyncio.sleep(10)
    await xknx.stop()


asyncio.run(main())



"""Unit test for KNX/IP TimerNotify objects."""

from xknx.knxip import KNXIPFrame, TimerNotify


class TestKNXIPTimerNotify:
    """Test class for KNX/IP TimerNotify objects."""

    def test_timer_notify(self) -> None:
        """Test parsing and streaming TimerNotify KNX/IP packet."""
        message_authentication_code = bytes.fromhex(
            "72 12 a0 3a aa e4 9d a8 56 89 77 4c 1d 2b 4d a4"
        )
        raw = (
            bytes.fromhex(
                "06 10 09 55 00 24"  # KNXnet/IP header - length 36 octets
                "c0 c1 c2 c3 c4 c5"  # timer value
                "00 fa 12 34 56 78"  # KNX serial number
                "af fe"  # message tag
            )
            + message_authentication_code
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, TimerNotify)
        assert knxipframe.body.timer_value == 211938428830917
        assert knxipframe.body.serial_number == bytes.fromhex("00 fa 12 34 56 78")
        assert knxipframe.body.message_tag == bytes.fromhex("af fe")
        assert (
            knxipframe.body.message_authentication_code == message_authentication_code
        )

        assert knxipframe.to_knx() == raw

        timer_notify = TimerNotify(
            timer_value=211938428830917,
            serial_number=bytes.fromhex("00 fa 12 34 56 78"),
            message_tag=bytes.fromhex("af fe"),
            message_authentication_code=message_authentication_code,
        )
        knxipframe2 = KNXIPFrame.init_from_body(timer_notify)

        assert knxipframe2.to_knx() == raw



"""Unit test for Sensor and ExposeSensor objects."""

from unittest.mock import AsyncMock, Mock, call

import pytest

from xknx import XKNX
from xknx.devices import BinarySensor, ExposeSensor, Sensor
from xknx.dpt import DPTArray, DPTBinary
from xknx.telegram import AddressFilter, Telegram, TelegramDirection
from xknx.telegram.address import GroupAddress, InternalGroupAddress
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite


class TestSensorExposeLoop:
    """Process incoming Telegrams and send the values to the bus again."""

    test_cases = [
        ("absolute_temperature", DPTArray((0x44, 0xD7, 0xD2, 0x87)), 1726.579),
        ("acceleration", DPTArray((0x45, 0x94, 0xD8, 0x5C)), 4763.045),
        ("acceleration_angular", DPTArray((0x45, 0xEA, 0x62, 0x33)), 7500.275),
        ("activation_energy", DPTArray((0x46, 0x0, 0x3E, 0xEE)), 8207.732),
        ("active_energy", DPTArray((0x26, 0x37, 0x49, 0x7F)), 641157503),
        ("active_energy_kwh", DPTArray((0x37, 0x5, 0x5, 0xEA)), 923076074),
        ("activity", DPTArray((0x45, 0x76, 0x0, 0xA4)), 3936.04),
        ("amplitude", DPTArray((0x45, 0x9A, 0xED, 0x8)), 4957.629),
        ("angle", DPTArray((0xE4,)), 322),
        ("angle_deg", DPTArray((0x44, 0x5C, 0x20, 0x2B)), 880.5026),
        ("angle_rad", DPTArray((0x44, 0x36, 0x75, 0x1)), 729.8282),
        ("angular_frequency", DPTArray((0x43, 0xBC, 0x20, 0x8D)), 376.2543),
        ("angular_momentum", DPTArray((0xC2, 0x75, 0xB7, 0xB5)), -61.4294),
        ("angular_velocity", DPTArray((0xC4, 0xD9, 0x10, 0xB4)), -1736.522),
        ("apparant_energy", DPTArray((0xD3, 0xBD, 0x1E, 0xA5)), -742580571),
        ("apparant_energy_kvah", DPTArray((0x49, 0x40, 0xC9, 0x9)), 1228982537),
        ("area", DPTArray((0x45, 0x63, 0x1E, 0xCD)), 3633.925),
        ("brightness", DPTArray((0xC3, 0x56)), 50006),
        ("capacitance", DPTArray((0x45, 0xC9, 0x1D, 0x9E)), 6435.702),
        ("charge_density_surface", DPTArray((0x45, 0xDB, 0x66, 0x9A)), 7020.825),
        ("charge_density_volume", DPTArray((0xC4, 0x8C, 0x33, 0xD7)), -1121.62),
        ("color_temperature", DPTArray((0x6C, 0x95)), 27797),
        ("common_temperature", DPTArray((0x45, 0xD9, 0xC6, 0x3F)), 6968.781),
        ("compressibility", DPTArray((0x45, 0x89, 0x94, 0xAA)), 4402.583),
        ("conductance", DPTArray((0x45, 0xA6, 0x28, 0xFA)), 5317.122),
        ("counter_pulses", DPTArray((0x9D,)), -99),
        ("current", DPTArray((0xCA, 0xCC)), 51916),
        ("delta_time_hrs", DPTArray((0x47, 0x80)), 18304),
        ("delta_time_min", DPTArray((0xB9, 0x7B)), -18053),
        ("delta_time_ms", DPTArray((0x58, 0x77)), 22647),
        ("delta_time_sec", DPTArray((0xA3, 0x6A)), -23702),
        ("density", DPTArray((0x44, 0xA5, 0xCB, 0x2B)), 1326.349),
        ("electrical_conductivity", DPTArray((0xC4, 0xC6, 0xF5, 0x71)), -1591.67),
        ("electric_charge", DPTArray((0x46, 0x14, 0xF6, 0xA0)), 9533.656),
        ("electric_current", DPTArray((0x45, 0xAD, 0x45, 0x8F)), 5544.695),
        ("electric_current_density", DPTArray((0x45, 0x7C, 0x57, 0xF8)), 4037.498),
        ("electric_dipole_moment", DPTArray((0x39, 0x01, 0x74, 0x2F)), 0.0001234568),
        ("electric_displacement", DPTArray((0xC5, 0x34, 0x8B, 0x02)), -2888.688),
        ("electric_field_strength", DPTArray((0xC6, 0x17, 0x1C, 0x39)), -9671.056),
        ("electric_flux", DPTArray((0x45, 0x8F, 0x6C, 0xFE)), 4589.624),
        ("electric_flux_density", DPTArray((0xC6, 0x0, 0x50, 0xA8)), -8212.164),
        ("electric_polarization", DPTArray((0x45, 0xF8, 0x89, 0xC7)), 7953.222),
        ("electric_potential", DPTArray((0xC6, 0x18, 0xA4, 0xAF)), -9769.171),
        ("electric_potential_difference", DPTArray((0xC6, 0xF, 0x1D, 0x6)), -9159.256),
        ("electromagnetic_moment", DPTArray((0x45, 0x82, 0x48, 0xAE)), 4169.085),
        ("electromotive_force", DPTArray((0x45, 0xBC, 0xEF, 0xEC)), 6045.99),
        ("energy", DPTArray((0x45, 0x4B, 0xB3, 0xF8)), 3259.248),
        ("enthalpy", DPTArray((0x76, 0xDD)), 287866.88),
        ("flow_rate_m3h", DPTArray((0x99, 0xEA, 0xC0, 0x55)), -1712668587),
        ("force", DPTArray((0x45, 0x9E, 0x2C, 0xE1)), 5061.61),
        ("frequency", DPTArray((0x45, 0xC2, 0x3C, 0x44)), 6215.533),
        ("heatcapacity", DPTArray((0xC5, 0xB3, 0x56, 0x7F)), -5738.812),
        ("heatflowrate", DPTArray((0x44, 0xEC, 0x80, 0x7B)), 1892.015),
        ("heat_quantity", DPTArray((0xC5, 0xA6, 0xB6, 0xD5)), -5334.854),
        ("humidity", DPTArray((0x7E, 0xE1)), 577044.48),
        ("impedance", DPTArray((0x45, 0xDD, 0x79, 0x6D)), 7087.178),
        ("illuminance", DPTArray((0x7C, 0x5E)), 366346.24),
        ("kelvin_per_percent", DPTArray((0xFA, 0xBD)), -441384.96),
        ("length", DPTArray((0xC5, 0x9D, 0xAE, 0xC5)), -5045.846),
        ("length_mm", DPTArray((0x56, 0xB9)), 22201),
        ("light_quantity", DPTArray((0x45, 0x4A, 0xF5, 0x68)), 3247.338),
        ("long_delta_timesec", DPTArray((0x45, 0xB2, 0x17, 0x54)), 1169299284),
        ("luminance", DPTArray((0x45, 0x18, 0xD9, 0x75)), 2445.591),
        ("luminous_flux", DPTArray((0x45, 0xBD, 0x16, 0x8)), 6050.754),
        ("luminous_intensity", DPTArray((0x46, 0xB, 0xBE, 0x7E)), 8943.623),
        ("magnetic_field_strength", DPTArray((0x44, 0x15, 0xF1, 0xAD)), 599.7762),
        ("magnetic_flux", DPTArray((0xC5, 0xCB, 0x3C, 0x98)), -6503.574),
        ("magnetic_flux_density", DPTArray((0x45, 0xB6, 0xBD, 0x42)), 5847.657),
        ("magnetic_moment", DPTArray((0xC3, 0x8E, 0x7F, 0x73)), -284.9957),
        ("magnetic_polarization", DPTArray((0x45, 0x8C, 0xFA, 0xCB)), 4511.349),
        ("magnetization", DPTArray((0x45, 0xF7, 0x9D, 0xA2)), 7923.704),
        ("magnetomotive_force", DPTArray((0xC6, 0x4, 0xC2, 0xDA)), -8496.713),
        ("mass", DPTArray((0x45, 0x8F, 0x70, 0xA4)), 4590.08),
        ("mass_flux", DPTArray((0xC6, 0x7, 0x34, 0xFF)), -8653.249),
        ("mol", DPTArray((0xC4, 0xA0, 0xF4, 0x6A)), -1287.638),
        ("momentum", DPTArray((0xC5, 0x27, 0xAA, 0x5A)), -2682.647),
        ("percent", DPTArray((0xE3,)), 89),
        ("percentU8", DPTArray((0x6B,)), 107),
        ("percentV8", DPTArray((0x20,)), 32),
        ("percentV16", DPTArray((0x8A, 0x2F)), -301.61),
        ("phaseanglerad", DPTArray((0x45, 0x54, 0xAC, 0x2D)), 3402.761),
        ("phaseangledeg", DPTArray((0xC5, 0x25, 0x13, 0x37)), -2641.201),
        ("power", DPTArray((0x45, 0xCB, 0xE2, 0x5C)), 6524.295),
        ("power_2byte", DPTArray((0x6D, 0x91)), 116736.00),
        ("power_density", DPTArray((0x65, 0x3E)), 54968.32),
        ("powerfactor", DPTArray((0xC5, 0x35, 0x28, 0x21)), -2898.508),
        ("ppm", DPTArray((0xF3, 0xC8)), -176947.20),
        ("pressure", DPTArray((0xC5, 0xE6, 0xE6, 0x62)), -7388.798),
        ("pressure_2byte", DPTArray((0x7C, 0xF4)), 415498.24),
        ("pulse", DPTArray((0xFC,)), 252),
        ("pulse_2byte_signed", DPTArray((0x80, 0x44)), -32700),
        ("rain_amount", DPTArray((0xF0, 0x1)), -335380.48),
        ("reactance", DPTArray((0x45, 0xB0, 0x50, 0x91)), 5642.071),
        ("reactive_energy", DPTArray((0x1A, 0x49, 0x6D, 0xA7)), 441019815),
        ("reactive_energy_kvarh", DPTArray((0xCC, 0x62, 0x5, 0x31)), -865991375),
        ("resistance", DPTArray((0xC5, 0xFC, 0x5F, 0xC3)), -8075.97),
        ("resistivity", DPTArray((0xC5, 0x57, 0x76, 0xC5)), -3447.423),
        ("rotation_angle", DPTArray((0x2D, 0xDC)), 11740),
        ("scene_number", DPTArray((0x1,)), 2),
        ("self_inductance", DPTArray((0xC4, 0xA1, 0xB0, 0x8)), -1293.501),
        ("solid_angle", DPTArray((0xC5, 0xC6, 0xE5, 0x48)), -6364.66),
        ("sound_intensity", DPTArray((0xC4, 0xF2, 0x56, 0xE9)), -1938.716),
        ("speed", DPTArray((0xC5, 0xCD, 0x1C, 0x6A)), -6563.552),
        ("stress", DPTArray((0x45, 0xDC, 0xA8, 0xF2)), 7061.118),
        ("surface_tension", DPTArray((0x46, 0xB, 0xAC, 0x11)), 8939.017),
        (
            "string",
            DPTArray(
                (
                    0x4B,
                    0x4E,
                    0x58,
                    0x20,
                    0x69,
                    0x73,
                    0x20,
                    0x4F,
                    0x4B,
                    0x0,
                    0x0,
                    0x0,
                    0x0,
                    0x0,
                )
            ),
            "KNX is OK",
        ),
        ("temperature", DPTArray((0x77, 0x88)), 315883.52),
        ("temperature_a", DPTArray((0xF1, 0xDB)), -257720.32),
        ("temperature_difference", DPTArray((0xC6, 0xC, 0x50, 0xBC)), -8980.184),
        ("temperature_difference_2byte", DPTArray((0xA9, 0xF4)), -495.36),
        ("temperature_f", DPTArray((0x67, 0xA9)), 80322.56),
        ("thermal_capacity", DPTArray((0x45, 0x83, 0xEA, 0xB2)), 4221.337),
        ("thermal_conductivity", DPTArray((0xC5, 0x9C, 0x4D, 0x23)), -5001.642),
        ("thermoelectric_power", DPTArray((0x41, 0xCF, 0x9E, 0x4F)), 25.9523),
        ("time_1", DPTArray((0x5E, 0x1E)), 32071.68),
        ("time_2", DPTArray((0xFB, 0x29)), -405995.52),
        ("time_period_100msec", DPTArray((0x6A, 0x35)), 2718900),
        ("time_period_10msec", DPTArray((0x32, 0x3)), 128030),
        ("time_period_hrs", DPTArray((0x29, 0xDE)), 10718),
        ("time_period_min", DPTArray((0x0, 0x54)), 84),
        ("time_period_msec", DPTArray((0x93, 0xC7)), 37831),
        ("time_period_sec", DPTArray((0xE0, 0xF5)), 57589),
        ("time_seconds", DPTArray((0x45, 0xEC, 0x91, 0x7D)), 7570.186),
        ("torque", DPTArray((0xC5, 0x9, 0x23, 0x60)), -2194.211),
        ("voltage", DPTArray((0x6D, 0xBF)), 120504.32),
        ("volume", DPTArray((0x46, 0x16, 0x98, 0x43)), 9638.065),
        ("volume_flow", DPTArray((0x7C, 0xF5)), 415825.92),
        ("volume_flux", DPTArray((0xC5, 0x4, 0x2D, 0x71)), -2114.84),
        ("weight", DPTArray((0x45, 0x20, 0x10, 0xE9)), 2561.057),
        ("work", DPTArray((0x45, 0x64, 0x5D, 0xBE)), 3653.859),
        ("wind_speed_ms", DPTArray((0x7D, 0x98)), 469237.76),
        ("wind_speed_kmh", DPTArray((0x7F, 0x55)), 615055.36),
        # # Generic DPT Without Min/Max and Unit.
        ("1byte_unsigned", DPTArray(0x08), 8),
        ("2byte_unsigned", DPTArray((0x30, 0x39)), 12345),
        ("2byte_signed", DPTArray((0x00, 0x01)), 1),
        ("2byte_float", DPTArray((0x2E, 0xA9)), 545.6),
        ("4byte_unsigned", DPTArray((0x00, 0x00, 0x00, 0x00)), 0),
        ("4byte_signed", DPTArray((0xFD, 0x1A, 0xA1, 0x09)), -48586487),
        ("4byte_float", DPTArray((0xC2, 0x09, 0xEE, 0xCC)), -34.4832),
    ]

    @pytest.mark.parametrize(("value_type", "test_payload", "test_value"), test_cases)
    async def test_array_sensor_loop(
        self, value_type: str, test_payload: DPTArray, test_value: float
    ) -> None:
        """Test sensor and expose_sensor with different values."""
        xknx = XKNX()
        xknx.cemi_handler = AsyncMock()
        await xknx.telegram_queue.start()

        expose = ExposeSensor(
            xknx,
            "TestExpose",
            group_address="1/1/1",
            value_type=value_type,
        )
        xknx.devices.async_add(expose)
        assert expose.resolve_state() is None
        # set a value from expose - HA sends strings for new values
        stringified_value = str(test_value)
        await expose.set(stringified_value)

        outgoing_telegram = Telegram(
            destination_address=GroupAddress("1/1/1"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueWrite(test_payload),
        )
        await xknx.telegrams.join()
        xknx.cemi_handler.send_telegram.assert_called_with(outgoing_telegram)
        assert expose.resolve_state() == test_value

        # init sensor after expose is set - with same group address
        sensor = Sensor(
            xknx,
            "TestSensor",
            group_address_state="1/1/1",
            value_type=value_type,
        )
        xknx.devices.async_add(sensor)
        assert sensor.resolve_state() is None

        # read sensor state (from expose as it has the same GA)
        # wait_for_result so we don't have to await self.xknx.telegrams.join()
        await sensor.sync(wait_for_result=True)
        read_telegram = Telegram(
            destination_address=GroupAddress("1/1/1"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueRead(),
        )
        response_telegram = Telegram(
            destination_address=GroupAddress("1/1/1"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueResponse(test_payload),
        )
        xknx.cemi_handler.send_telegram.assert_has_calls(
            [
                call(read_telegram),
                call(response_telegram),
            ]
        )
        # test if Sensor has successfully read from ExposeSensor
        assert sensor.resolve_state() == test_value
        assert expose.resolve_state() == sensor.resolve_state()
        await xknx.telegram_queue.stop()


class TestBinarySensorExposeLoop:
    """Process incoming Telegrams and send the values to the bus again."""

    @pytest.mark.parametrize(
        ("value_type", "test_payload", "test_value"),
        [
            ("binary", DPTBinary(0), False),
            ("binary", DPTBinary(1), True),
        ],
    )
    async def test_binary_sensor_loop(
        self, value_type: str, test_payload: DPTBinary, test_value: bool
    ) -> None:
        """Test binary_sensor and expose_sensor with binary values."""
        xknx = XKNX()
        xknx.cemi_handler = AsyncMock()
        await xknx.telegram_queue.start()

        expose = ExposeSensor(
            xknx,
            "TestExpose",
            group_address="1/1/1",
            value_type=value_type,
        )
        xknx.devices.async_add(expose)
        assert expose.resolve_state() is None

        await expose.set(test_value)
        await xknx.telegrams.join()
        outgoing_telegram = Telegram(
            destination_address=GroupAddress("1/1/1"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueWrite(test_payload),
        )
        xknx.cemi_handler.send_telegram.assert_called_with(outgoing_telegram)
        assert expose.resolve_state() == test_value

        bin_sensor = BinarySensor(
            xknx,
            "TestSensor",
            group_address_state="1/1/1",
        )
        xknx.devices.async_add(bin_sensor)
        assert bin_sensor.state is None

        # read sensor state (from expose as it has the same GA)
        # wait_for_result so we don't have to await self.xknx.telegrams.join()
        await bin_sensor.sync(wait_for_result=True)
        read_telegram = Telegram(
            destination_address=GroupAddress("1/1/1"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueRead(),
        )
        response_telegram = Telegram(
            destination_address=GroupAddress("1/1/1"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueResponse(test_payload),
        )
        xknx.cemi_handler.send_telegram.assert_has_calls(
            [
                call(read_telegram),
                call(response_telegram),
            ]
        )
        # test if Sensor has successfully read from ExposeSensor
        assert bin_sensor.state == test_value
        assert expose.resolve_state() == bin_sensor.state
        await xknx.telegram_queue.stop()


class TestBinarySensorInternalGroupAddressExposeLoop:
    """Process incoming Telegrams and send values to other devices."""

    @pytest.mark.parametrize(
        ("value_type", "test_payload", "test_value"),
        [
            ("binary", DPTBinary(0), False),
            ("binary", DPTBinary(1), True),
        ],
    )
    async def test_binary_sensor_loop(
        self, value_type: str, test_payload: DPTBinary, test_value: bool
    ) -> None:
        """Test binary_sensor and expose_sensor with binary values."""
        xknx = XKNX()
        xknx.cemi_handler = AsyncMock()

        telegram_callback = Mock()
        xknx.telegram_queue.register_telegram_received_cb(
            telegram_callback,
            address_filters=[AddressFilter("i-test")],
            match_for_outgoing=True,
        )
        await xknx.telegram_queue.start()

        expose = ExposeSensor(
            xknx,
            "TestExpose",
            group_address="i-test",
            value_type=value_type,
        )
        xknx.devices.async_add(expose)
        assert expose.resolve_state() is None

        await expose.set(test_value)
        await xknx.telegrams.join()
        outgoing_telegram = Telegram(
            destination_address=InternalGroupAddress("i-test"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueWrite(test_payload),
        )
        # InternalGroupAddress isn't passed to knxip_interface
        xknx.cemi_handler.send_telegram.assert_not_called()
        telegram_callback.assert_called_with(outgoing_telegram)
        assert expose.resolve_state() == test_value

        bin_sensor = BinarySensor(
            xknx,
            "TestSensor",
            group_address_state="i-test",
        )
        xknx.devices.async_add(bin_sensor)
        assert bin_sensor.state is None

        # read sensor state (from expose as it has the same GA)
        # wait_for_result so we don't have to await self.xknx.telegrams.join()
        await bin_sensor.sync(wait_for_result=True)
        read_telegram = Telegram(
            destination_address=InternalGroupAddress("i-test"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueRead(),
        )
        response_telegram = Telegram(
            destination_address=InternalGroupAddress("i-test"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueResponse(test_payload),
        )
        xknx.cemi_handler.send_telegram.assert_not_called()
        telegram_callback.assert_has_calls(
            [
                call(read_telegram),
                call(response_telegram),
            ]
        )
        # test if Sensor has successfully read from ExposeSensor
        assert bin_sensor.state == test_value
        assert expose.resolve_state() == bin_sensor.state
        await xknx.telegram_queue.stop()



"""Package for management procedures as described in KNX-Standard 3.5.2."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from xknx.exceptions import (
    ManagementConnectionError,
    ManagementConnectionRefused,
    ManagementConnectionTimeout,
)
from xknx.telegram import Telegram, apci, tpci
from xknx.telegram.address import (
    IndividualAddress,
    IndividualAddressableType,
)

if TYPE_CHECKING:
    from xknx import XKNX

logger = logging.getLogger("xknx.management.procedures")


async def dm_restart(xknx: XKNX, individual_address: IndividualAddressableType) -> None:
    """
    Restart the device.

    :param xknx: XKNX object
    :param individual_address: address of device to reset
    """
    async with xknx.management.connection(
        address=IndividualAddress(individual_address)
    ) as connection:
        logger.debug("Requesting a Basic Restart of %s.", individual_address)
        # A_Restart will not be ACKed by the device, so it is manually sent to avoid timeout and retry
        seq_num = next(connection.sequence_number)
        telegram = Telegram(
            destination_address=connection.address,
            source_address=xknx.current_address,
            payload=apci.Restart(),
            tpci=tpci.TDataConnected(sequence_number=seq_num),
        )
        await xknx.cemi_handler.send_telegram(telegram)


async def nm_individual_address_check(
    xknx: XKNX, individual_address: IndividualAddressableType
) -> bool:
    """
    Check if the individual address is occupied on the network.

    :param xknx: XKNX object
    :param individual_address: address to check
    """
    try:
        async with xknx.management.connection(
            address=IndividualAddress(individual_address)
        ) as connection:
            try:
                response = await connection.request(
                    payload=apci.DeviceDescriptorRead(descriptor=0),
                    expected=apci.DeviceDescriptorResponse,
                )

            except ManagementConnectionTimeout as ex:
                # if nothing is received (-> timeout) IA is free
                logger.debug("No device answered to connection attempt. %s", ex)
                return False
            if isinstance(response.payload, apci.DeviceDescriptorResponse):
                # if response is received IA is occupied
                logger.debug("Device found at %s", individual_address)
                return True
            return False
    except ManagementConnectionRefused as ex:
        # if Disconnect is received immediately, IA is occupied
        logger.debug("Device does not support transport layer connections. %s", ex)
        return True


async def nm_individual_address_read(
    xknx: XKNX,
    timeout: float | None = 3,
    raise_if_multiple: bool = False,
) -> list[IndividualAddress]:
    """
    Request individual addresses of all devices that are in programming mode.

    :param xknx: XKNX object
    :param timeout: specifies the timeout in seconds, the KNX specification requires a timeout of 3s
    :param raise_if_multiple: if true, ManagementConnectionError is raised when multiple devices are in programming mode
    :returns: list of individual address of devices in programming mode
    """

    addresses = []
    # initialize queue or event handler gathering broadcasts
    async with xknx.management.broadcast() as bc_context:
        await xknx.management.send_broadcast(apci.IndividualAddressRead())
        async for result in bc_context.receive(timeout=timeout):
            if isinstance(result.payload, apci.IndividualAddressResponse):
                addresses.append(result.source_address)
                if raise_if_multiple and (len(addresses) > 1):
                    raise ManagementConnectionError(
                        "More than one KNX device is in programming mode."
                    )
    return addresses


async def nm_individual_address_write(
    xknx: XKNX, individual_address: IndividualAddressableType
) -> None:
    """
    Write the individual address of a single device in programming mode.

    :param xknx: XKNX object
    :param individual_address: address to be written to KNX device
    """
    logger.debug("Writing individual address %s to device.", individual_address)

    # check if the address is already occupied on the network
    individual_address = IndividualAddress(individual_address)
    address_found = await nm_individual_address_check(xknx, individual_address)

    if address_found:
        logger.debug(
            "Individual address %s already present on the bus", individual_address
        )

    # check which devices are in programming mode
    dev_pgm_mode = await nm_individual_address_read(
        xknx, raise_if_multiple=True
    )  # raises exception if more than one device in programming mode
    if not dev_pgm_mode:
        logger.debug("No device in programming mode detected.")
        raise ManagementConnectionError("No device in programming mode detected.")

    # check if new and received addresses match
    if address_found:
        if individual_address != dev_pgm_mode[0]:
            logger.debug(
                "Device with address %s found and it is not in programming mode. Exiting to prevent address conflict.",
                individual_address,
            )
            raise ManagementConnectionError(
                f"A device was found with {individual_address}, cannot continue with programming."
            )
        # device in programming mode's address matches address that we want to write, so we can abort the operation safely
        logger.debug("Device already has requested address, no write operation needed.")
    else:
        await xknx.management.send_broadcast(
            payload=apci.IndividualAddressWrite(address=individual_address),
        )
        logger.debug("Wrote new address %s to device.", individual_address)

    async with xknx.management.connection(
        address=IndividualAddress(individual_address)
    ) as connection:
        logger.debug(
            "Checking if device exists at %s and restarting it.", individual_address
        )

        try:
            await connection.request(
                payload=apci.DeviceDescriptorRead(descriptor=0),
                expected=apci.DeviceDescriptorResponse,
            )
        except ManagementConnectionTimeout as ex:
            # if nothing is received (-> timeout) IA is free
            raise ManagementConnectionError(
                f"No device answered to connection attempt after write address operation. {ex}"
            ) from None

        logger.debug("Restating device, exiting programming mode.")
        # A_Restart will not be ACKed by the device, so it is manually sent to avoid timeout and retry
        seq_num = next(connection.sequence_number)
        telegram = Telegram(
            destination_address=connection.address,
            source_address=xknx.current_address,
            payload=apci.Restart(),
            tpci=tpci.TDataConnected(sequence_number=seq_num),
        )
        await xknx.cemi_handler.send_telegram(telegram)


# for backwards compatibility
nm_invididual_address_write = nm_individual_address_write


async def nm_individual_address_serial_number_read(
    xknx: XKNX,
    serial: bytes,
    timeout: float = 3,
) -> IndividualAddress | None:
    """Read individual address from device with specified serial number."""

    # initialize queue or event handler gathering broadcasts
    async with xknx.management.broadcast() as bc_context:
        await xknx.management.send_broadcast(
            payload=apci.IndividualAddressSerialRead(serial=serial)
        )
        async for result in bc_context.receive(timeout=timeout):
            if (
                isinstance(result.payload, apci.IndividualAddressSerialResponse)
                and result.payload.serial == serial
            ):
                return result.source_address

    return None


async def nm_individual_address_serial_number_write(
    xknx: XKNX, serial: bytes, individual_address: IndividualAddressableType
) -> None:
    """Write individual address to device with specified serial number."""
    individual_address = IndividualAddress(individual_address)
    await xknx.management.send_broadcast(
        payload=apci.IndividualAddressSerialWrite(
            address=individual_address,
            serial=serial,
        )
    )
    logger.debug(
        "Wrote new address %s to device with serial number %s.",
        individual_address,
        serial,
    )

    address = await nm_individual_address_serial_number_read(xknx=xknx, serial=serial)

    if address is None:
        raise ManagementConnectionError(f"No reply received from {serial!r}.")

    if address != individual_address:
        raise ManagementConnectionError(
            f"Failed to write serial address {individual_address} to device with serial {serial!r}. Detected {address}"
        )

    logger.debug(
        "New address %s validated on device with serial number %s.",
        individual_address,
        serial,
    )



"""Tests for KNX/IP routing indications using multicast."""

import asyncio
from typing import Final
from unittest.mock import Mock

from xknx import XKNX
from xknx.dpt.payload import DPTBinary
from xknx.io.connection import ConnectionConfig, ConnectionType
from xknx.telegram import GroupAddress, IndividualAddress, Telegram
from xknx.telegram.apci import GroupValueWrite
from xknx.telegram.telegram import TelegramDirection
from xknx.telegram.tpci import TDataGroup
from xknx.util import asyncio_timeout

# RFC 4727 multicast group for testing purposes
# to not mess with live installations when running tests
EXPERIMENTAL_MCAST_GRP: Final = "224.0.0.254"


async def test_routing_indication_multicast() -> None:
    """Test exchanging routing indication using multicast."""
    routing1_mock = Mock()
    data_received = asyncio.Event()
    received_data: Telegram | None = None

    def routing2_callback(telegram: Telegram) -> None:
        """Set test data from receiving xknx instance."""
        nonlocal received_data
        received_data = telegram
        data_received.set()

    msg = Telegram(
        destination_address=GroupAddress("1/2/2"),
        source_address=IndividualAddress("1.1.1"),
        payload=GroupValueWrite(DPTBinary(1)),
        tpci=TDataGroup(),
    )

    async with (
        XKNX(
            connection_config=ConnectionConfig(
                connection_type=ConnectionType.ROUTING,
                multicast_group=EXPERIMENTAL_MCAST_GRP,
            ),
            telegram_received_cb=routing1_mock,
        ) as xknx1,
        XKNX(
            connection_config=ConnectionConfig(
                connection_type=ConnectionType.ROUTING,
                multicast_group=EXPERIMENTAL_MCAST_GRP,
            ),
            telegram_received_cb=routing2_callback,
        ),
    ):
        await xknx1.telegrams.put(msg)
        async with asyncio_timeout(1):
            await data_received.wait()

        msg.direction = TelegramDirection.INCOMING
        routing1_mock.assert_not_called()
        assert received_data == msg



"""
Implementation of KNX 2 byte Float-values.

They correspond to the the following KDN DPT 9 class.
"""

from __future__ import annotations

from xknx.exceptions import ConversionError

from .dpt import DPTNumeric
from .payload import DPTArray, DPTBinary


class DPT2ByteFloat(DPTNumeric):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers.

    DPT 9.***
    """

    dpt_main_number = 9
    dpt_sub_number: int | None = None
    value_type = "2byte_float"
    payload_length = 2

    value_min = -671088.64
    value_max = 670760.96
    resolution = 0.01

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> float:
        """Parse/deserialize from KNX/IP raw data."""
        raw = cls.validate_payload(payload)
        data = (raw[0] * 256) + raw[1]
        exponent = (data >> 11) & 0x0F
        significand = data & 0x7FF
        sign = data >> 15

        if sign == 1:
            significand = significand - 2048

        value = float(significand << exponent) / 100

        if not cls._test_boundaries(value):
            raise ConversionError(f"Could not parse {cls.dpt_name()}", value=value)

        return value

    @classmethod
    def to_knx(cls, value: float) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        try:
            value = float(value)
            if not cls._test_boundaries(value):
                raise ValueError("Value out of range")

            knx_value = value * 100
            if round(knx_value) == 0:
                # ETS converts values near 0 like this
                # -0.0051 -> 0x87FF
                # -0.005 -> 0x8000  ... we don't want this since its equivalent to -20.48
                # 0.005 -> 0x0000
                # 0.0051 -> 0x0001
                return DPTArray((0x00, 0x00))

            exponent = 0
            while not -2048 <= knx_value <= 2047:
                exponent += 1
                knx_value /= 2

            mantisse = round(knx_value) & 0x7FF
            msb = exponent << 3 | mantisse >> 8
            if knx_value < 0:
                msb |= 0x80

            return DPTArray((msb, mantisse & 0xFF))
        except ValueError as err:
            raise ConversionError(
                f"Could not serialize {cls.dpt_name()}", value=value
            ) from err

    @classmethod
    def _test_boundaries(cls, value: float) -> bool:
        """Test if value is within defined range for this object."""
        return cls.value_min <= value <= cls.value_max


class DPTTemperature(DPT2ByteFloat):
    """DPT 9.001 DPT_Value_Temp."""

    dpt_main_number = 9
    dpt_sub_number = 1
    value_type = "temperature"
    unit = "°C"
    ha_device_class = "temperature"

    value_min = -273
    value_max = 670760


class DPTTemperatureDifference2Byte(DPT2ByteFloat):
    """DPT 9.002 DPT_Value_Tempd."""

    dpt_main_number = 9
    dpt_sub_number = 2
    value_type = "temperature_difference_2byte"
    unit = "K"
    ha_device_class = "temperature"

    value_min = -670760
    value_max = 670760


class DPTTemperatureA(DPT2ByteFloat):
    """DPT 9.003 DPT_Value_Tempa."""

    dpt_main_number = 9
    dpt_sub_number = 3
    value_type = "temperature_a"
    unit = "K/h"

    value_min = -670760
    value_max = 670760


class DPTLux(DPT2ByteFloat):
    """DPT 9.004 DPT_Value_Lux."""

    dpt_main_number = 9
    dpt_sub_number = 4
    value_type = "illuminance"
    unit = "lx"
    ha_device_class = "illuminance"

    value_min = 0
    value_max = 670760


class DPTWsp(DPT2ByteFloat):
    """DPT 9.005 DPT_Value_Ws Speed (m/s)."""

    dpt_main_number = 9
    dpt_sub_number = 5
    value_type = "wind_speed_ms"
    unit = "m/s"
    ha_device_class = "wind_speed"

    value_min = 0
    value_max = 670760


class DPTPressure2Byte(DPT2ByteFloat):
    """DPT 9.006 DPT_Value_Pres (Pa)."""

    dpt_main_number = 9
    dpt_sub_number = 6
    value_type = "pressure_2byte"
    unit = "Pa"
    ha_device_class = "pressure"

    value_min = 0
    value_max = 670760


class DPTHumidity(DPT2ByteFloat):
    """DPT 9.007 DPT_Value_Humidity."""

    dpt_main_number = 9
    dpt_sub_number = 7
    value_type = "humidity"
    unit = "%"
    ha_device_class = "humidity"

    value_min = 0
    value_max = 670760


class DPTPartsPerMillion(DPT2ByteFloat):
    """DPT 9.008 DPT_Value_parts/million."""

    dpt_main_number = 9
    dpt_sub_number = 8
    value_type = "ppm"
    unit = "ppm"


class DPTAirFlow(DPT2ByteFloat):
    """DPT 9.009 DPT_Value_AirFlow."""

    dpt_main_number = 9
    dpt_sub_number = 9
    value_type = "air_flow"
    unit = "m³/h"


class DPTTime1(DPT2ByteFloat):
    """DPT 9.010 DPT_Value_Time1 (s)."""

    dpt_main_number = 9
    dpt_sub_number = 10
    value_type = "time_1"
    unit = "s"

    value_min = -670760
    value_max = 670760


class DPTTime2(DPT2ByteFloat):
    """DPT 9.011 DPT_Value_Time2 (ms)."""

    dpt_main_number = 9
    dpt_sub_number = 11
    value_type = "time_2"
    unit = "ms"

    value_min = -670760
    value_max = 670760


class DPTVoltage(DPT2ByteFloat):
    """DPT 9.020 DPT_Value_Voltage."""

    dpt_main_number = 9
    dpt_sub_number = 20
    value_type = "voltage"
    unit = "mV"
    ha_device_class = "voltage"


class DPTCurrent(DPT2ByteFloat):
    """DPT 9.021 DPT_Value_Curr (mA)."""

    dpt_main_number = 9
    dpt_sub_number = 21
    value_type = "curr"
    unit = "mA"
    ha_device_class = "current"


class DPTPowerDensity(DPT2ByteFloat):
    """DPT 9.022 DPT_PowerDensity (W/m²)."""

    dpt_main_number = 9
    dpt_sub_number = 22
    value_type = "power_density"
    unit = "W/m²"


class DPTKelvinPerPercent(DPT2ByteFloat):
    """DPT 9.023 DPT_KelvinPerPercent (K/%)."""

    dpt_main_number = 9
    dpt_sub_number = 23
    value_type = "kelvin_per_percent"
    unit = "K/%"


class DPTPower2Byte(DPT2ByteFloat):
    """DPT 9.024 DPT_Power (kW)."""

    dpt_main_number = 9
    dpt_sub_number = 24
    value_type = "power_2byte"
    unit = "kW"
    ha_device_class = "power"


class DPTVolumeFlow(DPT2ByteFloat):
    """DPT 9.025 DPT_Value_Volume_Flow (L/h)."""

    dpt_main_number = 9
    dpt_sub_number = 25
    value_type = "volume_flow"
    unit = "L/h"


class DPTRainAmount(DPT2ByteFloat):
    """DPT 9.026 DPT_Rain_Amount (L/m²)."""

    dpt_main_number = 9
    dpt_sub_number = 26
    value_type = "rain_amount"
    unit = "L/m²"

    value_min = -671088.64
    value_max = 670760.96


class DPTTemperatureF(DPT2ByteFloat):
    """DPT 9.027 DPT_Value_Temp_F."""

    dpt_main_number = 9
    dpt_sub_number = 27
    value_type = "temperature_f"
    unit = "°F"
    ha_device_class = "temperature"

    value_min = -459.6
    value_max = 670760


class DPTWspKmh(DPT2ByteFloat):
    """DPT 9.028 DPT_Value_Wsp_kmh Speed (km/h)."""

    dpt_main_number = 9
    dpt_sub_number = 28
    value_type = "wind_speed_kmh"
    unit = "km/h"
    ha_device_class = "wind_speed"

    value_min = 0
    value_max = 670760


class DPTAbsoluteHumidity(DPT2ByteFloat):
    """DPT 9.029 DPT_Value_Absolute_Humidity."""

    dpt_main_number = 9
    dpt_sub_number = 29
    value_type = "absolute_humidity"
    unit = "g/m³"

    value_min = 0


class DPTConcentrationUGM3(DPT2ByteFloat):
    """DPT 9.030 DPT_Value_Concentration_μgm3."""

    dpt_main_number = 9
    dpt_sub_number = 30
    value_type = "concentration_ugm3"
    unit = "μg/m³"

    value_min = 0


class DPTEnthalpy(DPT2ByteFloat):
    """DPT 9.? 2-byte float value (with unit)."""

    dpt_main_number = 9
    # this is here for backwards compatibility
    dpt_sub_number = 60000
    value_type = "enthalpy"
    unit = "H"



"""
Implementation of KNX 4 byte Float-values.

They correspond to the the following KDN DPT 14 class.
"""

from __future__ import annotations

from math import ceil, log10
import struct
from typing import cast

from xknx.exceptions import ConversionError

from .dpt import DPTNumeric
from .payload import DPTArray, DPTBinary


class DPT4ByteFloat(DPTNumeric):
    """
    Abstraction for KNX 4 Octet Floating Point Numbers, with a maximum usable range as specified in IEEE 754.

    The largest positive finite float literal is 3.40282347e+38f.
    The smallest positive finite non-zero literal of type float is 1.40239846e-45f.
    The negative minimum finite float literal is -3.40282347e+38f.
    No value range are defined for DPTs 14.000-079.

    DPT 14.***
    """

    dpt_main_number = 14
    dpt_sub_number: int | None = None
    value_type = "4byte_float"
    payload_length = 4

    value_min = float("-inf")
    value_max = float("inf")
    resolution = 0.0000001

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> float:
        """Parse/deserialize from KNX/IP raw data (big endian)."""
        raw = cls.validate_payload(payload)
        try:
            raw_float = cast(float, struct.unpack(">f", bytes(raw))[0])
        except struct.error as err:
            raise ConversionError(f"Could not parse {cls.dpt_name()}", raw=raw) from err
        try:
            # round to 7 digit precision independent of exponent - same value as ETS 5.7 group monitor
            return round(raw_float, 7 - ceil(log10(abs(raw_float))))
        except (ValueError, OverflowError):
            # account for 0 and special values
            # ValueError: log10(0.0); ceil(float('nan'))
            # OverflowError: ceil(float('inf'))
            return raw_float

    @classmethod
    def to_knx(cls, value: float) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        try:
            knx_value = float(value)
            return DPTArray(struct.pack(">f", knx_value))
        except (ValueError, struct.error) as err:
            raise ConversionError(
                f"Could not serialize {cls.dpt_name()}", value=value
            ) from err


class DPTAcceleration(DPT4ByteFloat):
    """DPT 14.000 DPT_Value_Acceleration (ms-2)."""

    dpt_main_number = 14
    dpt_sub_number = 0
    value_type = "acceleration"
    unit = "m/s²"


class DPTAccelerationAngular(DPT4ByteFloat):
    """DPT 14.001 DPT_Value_Acceleration_Angular (rad s-2)."""

    dpt_main_number = 14
    dpt_sub_number = 1
    value_type = "acceleration_angular"
    unit = "rad/s²"


class DPTActivationEnergy(DPT4ByteFloat):
    """DPT 14.002 DPT_Value_Activation_Energy (J mol-1)."""

    dpt_main_number = 14
    dpt_sub_number = 2
    value_type = "activation_energy"
    unit = "J/mol"


class DPTActivity(DPT4ByteFloat):
    """DPT 14.003 DPT_Value_Activity (s-1)."""

    dpt_main_number = 14
    dpt_sub_number = 3
    value_type = "activity"
    unit = "s⁻¹"


class DPTMol(DPT4ByteFloat):
    """DPT 14.004 DPT_Value_Mol (mol)."""

    dpt_main_number = 14
    dpt_sub_number = 4
    value_type = "mol"
    unit = "mol"


class DPTAmplitude(DPT4ByteFloat):
    """DPT 14.005 DPT_Value_Amplitude."""

    dpt_main_number = 14
    dpt_sub_number = 5
    value_type = "amplitude"


class DPTAngleRad(DPT4ByteFloat):
    """DPT 14.006 DPT_Value_AngleRad (rad)."""

    dpt_main_number = 14
    dpt_sub_number = 6
    value_type = "angle_rad"
    unit = "rad"


class DPTAngleDeg(DPT4ByteFloat):
    """DPT 14.007 DPT_Value_AngleDeg ((degree))."""

    dpt_main_number = 14
    dpt_sub_number = 7
    value_type = "angle_deg"
    unit = "°"


class DPTAngularMomentum(DPT4ByteFloat):
    """DPT 14.008 DPT_Value_Angular_Momentum (J s)."""

    dpt_main_number = 14
    dpt_sub_number = 8
    value_type = "angular_momentum"
    unit = "J s"


class DPTAngularVelocity(DPT4ByteFloat):
    """DPT 14.009 DPT_Value_Angular_Velocity."""

    dpt_main_number = 14
    dpt_sub_number = 9
    value_type = "angular_velocity"
    unit = "rad/s"


class DPTArea(DPT4ByteFloat):
    """DPT 14.010 DPT_Value_Area."""

    dpt_main_number = 14
    dpt_sub_number = 10
    value_type = "area"
    unit = "m²"


class DPTCapacitance(DPT4ByteFloat):
    """DPT 14.011 DPT_Value_Capacitance."""

    dpt_main_number = 14
    dpt_sub_number = 11
    value_type = "capacitance"
    unit = "F"


class DPTChargeDensitySurface(DPT4ByteFloat):
    """DPT 14.012 DPT_Value_Charge_DensitySurface."""

    dpt_main_number = 14
    dpt_sub_number = 12
    value_type = "charge_density_surface"
    unit = "C/m²"


class DPTChargeDensityVolume(DPT4ByteFloat):
    """DPT 14.013 DPT_Value_Charge_DensityVolume."""

    dpt_main_number = 14
    dpt_sub_number = 13
    value_type = "charge_density_volume"
    unit = "C/m³"


class DPTCompressibility(DPT4ByteFloat):
    """DPT 14.014 DPT_Value_Compressibility."""

    dpt_main_number = 14
    dpt_sub_number = 14
    value_type = "compressibility"
    unit = "m²/N"


class DPTConductance(DPT4ByteFloat):
    """DPT 14.015 DPT_Value_Conductance."""

    dpt_main_number = 14
    dpt_sub_number = 15
    value_type = "conductance"
    unit = "S"


class DPTElectricalConductivity(DPT4ByteFloat):
    """DPT 14.016 DPT_Value_Electrical_Conductivity."""

    dpt_main_number = 14
    dpt_sub_number = 16
    value_type = "electrical_conductivity"
    unit = "S/m"


class DPTDensity(DPT4ByteFloat):
    """DPT 14.017 DPT_Value_Density."""

    dpt_main_number = 14
    dpt_sub_number = 17
    value_type = "density"
    unit = "kg/m³"


class DPTElectricCharge(DPT4ByteFloat):
    """DPT 14.018 DPT_Value_Electric_Charge."""

    dpt_main_number = 14
    dpt_sub_number = 18
    value_type = "electric_charge"
    unit = "C"


class DPTElectricCurrent(DPT4ByteFloat):
    """DPT 14.019 DPT_Value_Electric_Current."""

    dpt_main_number = 14
    dpt_sub_number = 19
    value_type = "electric_current"
    unit = "A"
    ha_device_class = "current"


class DPTElectricCurrentDensity(DPT4ByteFloat):
    """DPT 14.020 DPT_Value_Electric_CurrentDensity."""

    dpt_main_number = 14
    dpt_sub_number = 20
    value_type = "electric_current_density"
    unit = "A/m²"


class DPTElectricDipoleMoment(DPT4ByteFloat):
    """DPT 14.021 DPT_Value_Electric_DipoleMoment."""

    dpt_main_number = 14
    dpt_sub_number = 21
    value_type = "electric_dipole_moment"
    unit = "C m"


class DPTElectricDisplacement(DPT4ByteFloat):
    """DPT 14.022 DPT_Value_Electric_Displacement."""

    dpt_main_number = 14
    dpt_sub_number = 22
    value_type = "electric_displacement"
    unit = "C/m²"


class DPTElectricFieldStrength(DPT4ByteFloat):
    """DPT 14.023 DPT_Value_Electric_FieldStrength."""

    dpt_main_number = 14
    dpt_sub_number = 23
    value_type = "electric_field_strength"
    unit = "V/m"


class DPTElectricFlux(DPT4ByteFloat):
    """DPT 14.024 DPT_Value_Electric_Flux."""

    dpt_main_number = 14
    dpt_sub_number = 24
    value_type = "electric_flux"
    unit = "c"


class DPTElectricFluxDensity(DPT4ByteFloat):
    """DPT 14.025 DPT_Value_Electric_FluxDensity."""

    dpt_main_number = 14
    dpt_sub_number = 25
    value_type = "electric_flux_density"
    unit = "C/m²"


class DPTElectricPolarization(DPT4ByteFloat):
    """DPT 14.026 DPT_Value_Electric_Polarization."""

    dpt_main_number = 14
    dpt_sub_number = 26
    value_type = "electric_polarization"
    unit = "C/m²"


class DPTElectricPotential(DPT4ByteFloat):
    """DPT 14.027 DPT_Value_Electric_Potential."""

    dpt_main_number = 14
    dpt_sub_number = 27
    value_type = "electric_potential"
    unit = "V"


class DPTElectricPotentialDifference(DPT4ByteFloat):
    """DPT 14.028 DPT_Value_Electric_PotentialDifference."""

    dpt_main_number = 14
    dpt_sub_number = 28
    value_type = "electric_potential_difference"
    unit = "V"


class DPTElectromagneticMoment(DPT4ByteFloat):
    """DPT 14.029 DPT_Value_ElectromagneticMoment."""

    dpt_main_number = 14
    dpt_sub_number = 29
    value_type = "electromagnetic_moment"
    unit = "A m²"


class DPTElectromotiveForce(DPT4ByteFloat):
    """DPT 14.030 DPT_Value_Electromotive_Force."""

    dpt_main_number = 14
    dpt_sub_number = 30
    value_type = "electromotive_force"
    unit = "V"


class DPTEnergy(DPT4ByteFloat):
    """DPT 14.031 DPT_Value_Energy."""

    dpt_main_number = 14
    dpt_sub_number = 31
    value_type = "energy"
    unit = "J"


class DPTForce(DPT4ByteFloat):
    """DPT 14.032 DPT_Value_Force."""

    dpt_main_number = 14
    dpt_sub_number = 32
    value_type = "force"
    unit = "N"


class DPTFrequency(DPT4ByteFloat):
    """DPT 14.033 DPT_Value_Frequency."""

    dpt_main_number = 14
    dpt_sub_number = 33
    value_type = "frequency"
    unit = "Hz"
    ha_device_class = "frequency"


class DPTAngularFrequency(DPT4ByteFloat):
    """DPT 14.034 DPT_Value_Angular_Frequency."""

    dpt_main_number = 14
    dpt_sub_number = 34
    value_type = "angular_frequency"
    unit = "rad/s"


class DPTHeatCapacity(DPT4ByteFloat):
    """DPT 14.035 DPT_Value_Heat_Capacity."""

    dpt_main_number = 14
    dpt_sub_number = 35
    value_type = "heatcapacity"
    unit = "J/K"


class DPTHeatFlowRate(DPT4ByteFloat):
    """DPT 14.036 DPT_Value_Heat_Flow_Rate."""

    dpt_main_number = 14
    dpt_sub_number = 36
    value_type = "heatflowrate"
    unit = "W"


class DPTHeatQuantity(DPT4ByteFloat):
    """DPT 14.037 DPT_Value_Heat_Quantity."""

    dpt_main_number = 14
    dpt_sub_number = 37
    value_type = "heat_quantity"
    unit = "J"


class DPTImpedance(DPT4ByteFloat):
    """DPT 14.038 DPT_Value_Impedance."""

    dpt_main_number = 14
    dpt_sub_number = 38
    value_type = "impedance"
    unit = "Ω"


class DPTLength(DPT4ByteFloat):
    """DPT 14.039 DPT_Value_Length."""

    dpt_main_number = 14
    dpt_sub_number = 39
    value_type = "length"
    unit = "m"
    ha_device_class = "distance"


class DPTLightQuantity(DPT4ByteFloat):
    """DPT 14.040 DPT_Value_Light_Quantity."""

    dpt_main_number = 14
    dpt_sub_number = 40
    value_type = "light_quantity"
    unit = "lm s"


class DPTLuminance(DPT4ByteFloat):
    """DPT 14.041 DPT_Value_Luminance."""

    dpt_main_number = 14
    dpt_sub_number = 41
    value_type = "luminance"
    unit = "cd/m²"


class DPTLuminousFlux(DPT4ByteFloat):
    """DPT 14.042 DPT_Value_Heat_Flow_Rate."""

    dpt_main_number = 14
    dpt_sub_number = 42
    value_type = "luminous_flux"
    unit = "lm"


class DPTLuminousIntensity(DPT4ByteFloat):
    """DPT 14.043 DPT_Value_Luminous_Intensity."""

    dpt_main_number = 14
    dpt_sub_number = 43
    value_type = "luminous_intensity"
    unit = "cd"


class DPTMagneticFieldStrength(DPT4ByteFloat):
    """DPT 14.044 DPT_Value_Magnetic_FieldStrength."""

    dpt_main_number = 14
    dpt_sub_number = 44
    value_type = "magnetic_field_strength"
    unit = "A/m"


class DPTMagneticFlux(DPT4ByteFloat):
    """DPT 14.045 DPT_Value_Magnetic_Flux."""

    dpt_main_number = 14
    dpt_sub_number = 45
    value_type = "magnetic_flux"
    unit = "Wb"


class DPTMagneticFluxDensity(DPT4ByteFloat):
    """DPT 14.046 DPT_Value_Magnetic_FluxDensity."""

    dpt_main_number = 14
    dpt_sub_number = 46
    value_type = "magnetic_flux_density"
    unit = "T"


class DPTMagneticMoment(DPT4ByteFloat):
    """DPT 14.047 DPT_Value_Magnetic_Moment."""

    dpt_main_number = 14
    dpt_sub_number = 47
    value_type = "magnetic_moment"
    unit = "A m²"


class DPTMagneticPolarization(DPT4ByteFloat):
    """DPT 14.048 DPT_Value_Magnetic_Polarization."""

    dpt_main_number = 14
    dpt_sub_number = 48
    value_type = "magnetic_polarization"
    unit = "T"


class DPTMagnetization(DPT4ByteFloat):
    """DPT 14.049 DPT_Value_Magnetization."""

    dpt_main_number = 14
    dpt_sub_number = 49
    value_type = "magnetization"
    unit = "A/m"


class DPTMagnetomotiveForce(DPT4ByteFloat):
    """DPT 14.050 DPT_Value_MagnetomotiveForce."""

    dpt_main_number = 14
    dpt_sub_number = 50
    value_type = "magnetomotive_force"
    unit = "A"


class DPTMass(DPT4ByteFloat):
    """DPT 14.051 DPT_Value_Mass."""

    dpt_main_number = 14
    dpt_sub_number = 51
    value_type = "mass"
    unit = "kg"
    ha_device_class = "weight"


class DPTMassFlux(DPT4ByteFloat):
    """DPT 14.052 DPT_Value_MassFlux."""

    dpt_main_number = 14
    dpt_sub_number = 52
    value_type = "mass_flux"
    unit = "kg/s"


class DPTMomentum(DPT4ByteFloat):
    """DPT 14.053 DPT_Value_Momentum."""

    dpt_main_number = 14
    dpt_sub_number = 53
    value_type = "momentum"
    unit = "N/s"


class DPTPhaseAngleRad(DPT4ByteFloat):
    """DPT 14.054 DPT_Value_Phase_Angle, Radiant."""

    dpt_main_number = 14
    dpt_sub_number = 54
    value_type = "phaseanglerad"
    unit = "rad"


class DPTPhaseAngleDeg(DPT4ByteFloat):
    """DPT 14.055 DPT_Value_Phase_Angle, Degree."""

    dpt_main_number = 14
    dpt_sub_number = 55
    value_type = "phaseangledeg"
    unit = "°"


class DPTPower(DPT4ByteFloat):
    """DPT 14.056 DPT_Value_Power."""

    dpt_main_number = 14
    dpt_sub_number = 56
    value_type = "power"
    unit = "W"
    ha_device_class = "power"


class DPTPowerFactor(DPT4ByteFloat):
    """DPT 14.057 DPT_Value_Power."""

    dpt_main_number = 14
    dpt_sub_number = 57
    value_type = "powerfactor"
    ha_device_class = "power_factor"


class DPTPressure(DPT4ByteFloat):
    """DPT 14.058 DPT_Value_Pressure."""

    dpt_main_number = 14
    dpt_sub_number = 58
    value_type = "pressure"
    unit = "Pa"
    ha_device_class = "pressure"


class DPTReactance(DPT4ByteFloat):
    """DPT 14.059 DPT_Value_Reactance."""

    dpt_main_number = 14
    dpt_sub_number = 59
    value_type = "reactance"
    unit = "Ω"


class DPTResistance(DPT4ByteFloat):
    """DPT 14.060 DPT_Value_Resistance."""

    dpt_main_number = 14
    dpt_sub_number = 60
    value_type = "resistance"
    unit = "Ω"


class DPTResistivity(DPT4ByteFloat):
    """DPT 14.061 DPT_Value_Resistivity."""

    dpt_main_number = 14
    dpt_sub_number = 61
    value_type = "resistivity"
    unit = "Ωm"


class DPTSelfInductance(DPT4ByteFloat):
    """DPT 14.062 DPT_Value_SelfInductance."""

    dpt_main_number = 14
    dpt_sub_number = 62
    value_type = "self_inductance"
    unit = "H"


class DPTSolidAngle(DPT4ByteFloat):
    """DPT 14.063 DPT_Value_SolidAngle."""

    dpt_main_number = 14
    dpt_sub_number = 63
    value_type = "solid_angle"
    unit = "sr"


class DPTSoundIntensity(DPT4ByteFloat):
    """DPT 14.064 DPT_Value_Sound_Intensity."""

    dpt_main_number = 14
    dpt_sub_number = 64
    value_type = "sound_intensity"
    unit = "W/m²"


class DPTSpeed(DPT4ByteFloat):
    """DPT 14.065 DPT_Value_Speed."""

    dpt_main_number = 14
    dpt_sub_number = 65
    value_type = "speed"
    unit = "m/s"
    ha_device_class = "speed"


class DPTStress(DPT4ByteFloat):
    """DPT 14.066 DPT_Value_Stress."""

    dpt_main_number = 14
    dpt_sub_number = 66
    value_type = "stress"
    unit = "Pa"


class DPTSurfaceTension(DPT4ByteFloat):
    """DPT 14.067 DPT_Value_Surface_Tension."""

    dpt_main_number = 14
    dpt_sub_number = 67
    value_type = "surface_tension"
    unit = "N/m"


class DPTCommonTemperature(DPT4ByteFloat):
    """DPT 14.068 DPT_Value_Common_Temperature."""

    dpt_main_number = 14
    dpt_sub_number = 68
    value_type = "common_temperature"
    unit = "°C"


class DPTAbsoluteTemperature(DPT4ByteFloat):
    """DPT 14.069 DPT_Value_Absolute_Temperature."""

    dpt_main_number = 14
    dpt_sub_number = 69
    value_type = "absolute_temperature"
    unit = "K"


class DPTTemperatureDifference(DPT4ByteFloat):
    """DPT 14.070 DPT_Value_TemperatureDifference."""

    dpt_main_number = 14
    dpt_sub_number = 70
    value_type = "temperature_difference"
    unit = "K"


class DPTThermalCapacity(DPT4ByteFloat):
    """DPT 14.071 DPT_Value_Thermal_Capacity."""

    dpt_main_number = 14
    dpt_sub_number = 71
    value_type = "thermal_capacity"
    unit = "J/K"


class DPTThermalConductivity(DPT4ByteFloat):
    """DPT 14.072 DPT_Value_Thermal_Conductivity."""

    dpt_main_number = 14
    dpt_sub_number = 72
    value_type = "thermal_conductivity"
    unit = "W/mK"


class DPTThermoelectricPower(DPT4ByteFloat):
    """DPT 14.073 DPT_Value_ThermoelectricPower."""

    dpt_main_number = 14
    dpt_sub_number = 73
    value_type = "thermoelectric_power"
    unit = "V/K"


class DPTTimeSeconds(DPT4ByteFloat):
    """DPT 14.074 DPT_Value_Time."""

    dpt_main_number = 14
    dpt_sub_number = 74
    value_type = "time_seconds"
    unit = "s"


class DPTTorque(DPT4ByteFloat):
    """DPT 14.075 DPT_Value_Torque."""

    dpt_main_number = 14
    dpt_sub_number = 75
    value_type = "torque"
    unit = "Nm"


class DPTVolume(DPT4ByteFloat):
    """DPT 14.076 DPT_Value_Volume."""

    dpt_main_number = 14
    dpt_sub_number = 76
    value_type = "volume"
    unit = "m³"


class DPTVolumeFlux(DPT4ByteFloat):
    """DPT 14.077 DPT_Value_Volume_Flux."""

    dpt_main_number = 14
    dpt_sub_number = 77
    value_type = "volume_flux"
    unit = "m³/s"


class DPTWeight(DPT4ByteFloat):
    """DPT 14.078 DPT_Value_Weight."""

    dpt_main_number = 14
    dpt_sub_number = 78
    value_type = "weight"
    unit = "N"


class DPTWork(DPT4ByteFloat):
    """DPT 14.079 DPT_Value_Work."""

    dpt_main_number = 14
    dpt_sub_number = 79
    value_type = "work"
    unit = "J"


class DPTApparentPower(DPT4ByteFloat):
    """DPT 14.080 DPT_Value_Apparent_Power."""

    dpt_main_number = 14
    dpt_sub_number = 80
    value_type = "apparent_power"
    unit = "VA"
    ha_device_class = "apparent_power"


class DPTVolumeFluxMeter(DPT4ByteFloat):
    """DPT 14.1200 DPT_Volume_Flux_Meter (volume flux for meters)."""

    dpt_main_number = 14
    dpt_sub_number = 1200
    value_type = "volume_flux_meter"
    unit = "m³/h"


class DPTVolumeFluxLs(DPT4ByteFloat):
    """DPT 14.1201 DPT_Volume_Flux_ls (volume flux for meters)."""

    dpt_main_number = 14
    dpt_sub_number = 1201
    value_type = "volume_flux_ls"
    unit = "L/s"



"""Unit test for RemoteValueDptValue1Ucount objects."""

import pytest

from xknx import XKNX
from xknx.dpt import DPTArray, DPTBinary
from xknx.exceptions import ConversionError
from xknx.remote_value import RemoteValueDptValue1Ucount
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestRemoteValueDptValue1Ucount:
    """Test class for RemoteValueDptValue1Ucount objects."""

    def test_to_knx(self) -> None:
        """Test to_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueDptValue1Ucount(xknx)
        assert remote_value.to_knx(10) == DPTArray((0x0A,))

    def test_from_knx(self) -> None:
        """Test from_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueDptValue1Ucount(xknx)
        assert remote_value.from_knx(DPTArray((0x0A,))) == 10

    def test_to_knx_error(self) -> None:
        """Test to_knx function with wrong parametern."""
        xknx = XKNX()
        remote_value = RemoteValueDptValue1Ucount(xknx)
        with pytest.raises(ConversionError):
            remote_value.to_knx(256)
        with pytest.raises(ConversionError):
            remote_value.to_knx("256")

    async def test_set(self) -> None:
        """Test setting value."""
        xknx = XKNX()
        remote_value = RemoteValueDptValue1Ucount(
            xknx, group_address=GroupAddress("1/2/3")
        )
        remote_value.set(10)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x0A,))),
        )
        remote_value.set(11)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x0B,))),
        )

    def test_process(self) -> None:
        """Test process telegram."""
        xknx = XKNX()
        remote_value = RemoteValueDptValue1Ucount(
            xknx, group_address=GroupAddress("1/2/3")
        )
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x0A,))),
        )
        remote_value.process(telegram)
        assert remote_value.value == 10

    def test_to_process_error(self) -> None:
        """Test process erroneous telegram."""
        xknx = XKNX()
        remote_value = RemoteValueDptValue1Ucount(
            xknx, group_address=GroupAddress("1/2/3")
        )

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        assert remote_value.process(telegram) is False

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65))),
        )
        assert remote_value.process(telegram) is False

        assert remote_value.value is None



"""Example for connecting to a specific KNX interface."""

import asyncio
import logging
import time

from xknx import XKNX
from xknx.io import ConnectionConfig, ConnectionType
from xknx.tools import read_group_value

logging.basicConfig(level=logging.INFO)
logging.getLogger("xknx.log").level = logging.DEBUG
logging.getLogger("xknx.knx").level = logging.DEBUG


async def main() -> None:
    """Connect to specific tunnelling server, time a request for a group address value."""
    connection_config = ConnectionConfig(
        connection_type=ConnectionType.TUNNELING,
        gateway_ip="10.1.0.40",
        # local_ip="10.1.0.123",
        # route_back=True,
    )
    xknx = XKNX(connection_config=connection_config)

    async with xknx:
        start_time = time.time()
        result = await read_group_value(xknx, "5/1/20", value_type="temperature")
        print(f"Value: {result} - took {(time.time() - start_time):0.3f} seconds")


asyncio.run(main())



"""
Module for Serialization and Deserialization of KNX Search Requests.

Search Requests are used to search for KNX/IP devices within the network.
"""

from __future__ import annotations

from .body import KNXIPBody
from .hpai import HPAI
from .knxip_enum import KNXIPServiceType


class SearchRequest(KNXIPBody):
    """Representation of a KNX Search Request."""

    SERVICE_TYPE = KNXIPServiceType.SEARCH_REQUEST

    def __init__(self, discovery_endpoint: HPAI | None = None) -> None:
        """Initialize SearchRequest object."""
        self.discovery_endpoint = discovery_endpoint or HPAI()

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return HPAI.LENGTH

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        return self.discovery_endpoint.from_knx(raw)

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return self.discovery_endpoint.to_knx()

    def __repr__(self) -> str:
        """Return object as readable string."""
        return f'<SearchRequest discovery_endpoint="{self.discovery_endpoint}" />'
